#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import zlib
import base64
import datetime
import traceback
import contextlib

import netius.common

from netius.common import PLAIN_ENCODING, CHUNKED_ENCODING,\
    GZIP_ENCODING, DEFLATE_ENCODING

Z_PARTIAL_FLUSH = 1
""" The zlib constant value representing the partial flush
of the current zlib stream, this value has to be defined
locally as it is not defines under the zlib module """

ENCODING_MAP = dict(
    plain = PLAIN_ENCODING,
    chunked = CHUNKED_ENCODING,
    gzip = GZIP_ENCODING,
    deflate = DEFLATE_ENCODING
)
""" The map associating the various types of encoding with
the corresponding integer value for each of them this is used
in the initial construction of the server """

class HTTPConnection(netius.Connection):

    def __init__(self, encoding = PLAIN_ENCODING, *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.encoding = encoding
        self.current = encoding
        self.parser = None
        self.legacy = True
        self.gzip_m = dict()

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.HTTPParser(
            self,
            type = netius.common.REQUEST,
            store = True
        )
        self.parser.bind("on_data", self.on_data)

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.parser: self.parser.destroy()
        if self.gzip_m: self._close_gzip(safe = True)

    def info_dict(self, full = False):
        info = netius.Connection.info_dict(self, full = full)
        info.update(
            encoding = self.encoding,
            current = self.current
        )
        if full: info.update(
            parser = self.parser.info_dict()
        )
        return info

    def flush(self, stream = None, callback = None):
        encoding = min(self.current, CHUNKED_ENCODING) if\
            self.owner.dynamic else self.current

        if encoding == DEFLATE_ENCODING:
            self._flush_gzip(stream = stream, callback = callback)
        elif encoding == GZIP_ENCODING:
            self._flush_gzip(stream = stream, callback = callback)
        elif encoding == CHUNKED_ENCODING:
            self._flush_chunked(stream = stream, callback = callback)
        elif encoding == PLAIN_ENCODING:
            self._flush_plain(stream = stream, callback = callback)

        self.current = self.encoding

        self.owner.on_flush_http(
            self.connection_ctx,
            self.parser_ctx,
            encoding = encoding
        )

    def flush_s(self, stream = None, callback = None):
        return self.flush(stream = stream, callback = callback)

    def send_base(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        data = netius.legacy.bytes(data) if data else data
        encoding = min(self.current, CHUNKED_ENCODING) if\
            self.owner.dynamic else self.current

        if encoding == PLAIN_ENCODING:
            return self.send_plain(
                data,
                stream = stream,
                final = final,
                delay = delay,
                callback = callback
            )
        elif encoding == CHUNKED_ENCODING:
            return self.send_chunked(
                data,
                stream = stream,
                final = final,
                delay = delay,
                callback = callback
            )
        elif encoding == GZIP_ENCODING:
            return self.send_gzip(
                data,
                stream = stream,
                final = final,
                delay = delay,
                callback = callback
            )
        elif encoding == DEFLATE_ENCODING:
            return self.send_gzip(
                data,
                stream = stream,
                final = final,
                delay = delay,
                callback = callback
            )

    def send_plain(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        return self.send(
            data,
            delay = delay,
            callback = callback
        )

    def send_chunked(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None
    ):
        # in case there's no valid data to be sent uses the plain
        # send method to send the empty string and returns immediately
        # to the caller method, to avoid any problems
        if not data: return self.send_plain(
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

        # creates the new list that is going to be used to store
        # the various parts of the chunk and then calculates the
        # size (in bytes) of the data that is going to be sent
        buffer = []
        size = len(data)

        # creates the various parts of the chunk with the size
        # of the data that is going to be sent and then adds
        # each of the parts to the chunk buffer list
        buffer.append(netius.legacy.bytes("%x\r\n" % size))
        buffer.append(data)
        buffer.append(netius.legacy.bytes("\r\n"))

        # joins the buffer containing the chunk parts and then
        # sends it to the connection using the plain method
        buffer_s = b"".join(buffer)
        return self.send_plain(
            buffer_s,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

    def send_gzip(
        self,
        data,
        stream = None,
        final = True,
        delay = True,
        callback = None,
        level = 6
    ):
        # verifies if the provided data buffer is valid and in
        # in case it's not propagates the sending to the upper
        # layer (chunked sending) for proper processing
        if not data: return self.send_chunked(
            data,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

        # tries to retrieve the gzip object for the current stream
        # in case it's a new connection one will be created with the
        # appropriate compression level (as expected)
        gzip = self._get_gzip(stream, level = level)

        # compresses the provided data string and removes the
        # initial data contents of the compressed data because
        # they are not part of the gzip specification, notice
        # that in case the resulting of the compress operation
        # is not valid a sync flush operation is performed
        data_c = gzip.compress(data)
        if not data_c: data_c = gzip.flush(Z_PARTIAL_FLUSH)

        # sends the compressed data to the client endpoint setting
        # the correct callback values as requested
        return self.send_chunked(
            data_c,
            stream = stream,
            final = final,
            delay = delay,
            callback = callback
        )

    def send_response(
        self,
        data = None,
        headers = None,
        version = None,
        code = 200,
        code_s = None,
        apply = False,
        stream = None,
        final = True,
        flush = True,
        delay = True,
        callback = None
    ):
        # retrieves the various parts that define the response
        # and runs a series of normalization processes to retrieve
        # the relevant information of the data to be sent to client
        data = data or b""
        data = netius.legacy.bytes(data)
        headers = headers or dict()
        data_l = len(data) if data else 0
        is_empty = code in (204, 304) and data_l == 0

        # runs a series of verifications taking into account the type
        # of the method defined in the current request, for instance if
        # the current request is a HEAD one then no data is sent (as expected)
        if self.parser_ctx.method and self.parser_ctx.method.upper() == "HEAD":
            data = b""

        # verifies if the content length header is currently present
        # in the provided headers and in case it's not inserts it
        if not "content-length" in headers and not is_empty:
            headers["content-length"] = str(data_l)

        # in case the apply flag is set the apply all operation is performed
        # so that a series of headers are applied to the current context
        # (things like the name of the server connection, etc)
        if apply: self.owner._apply_all(self.parser, self, headers)

        # sends the initial headers data (including status line), this should
        # trigger the initial data sent to the peer/client
        count = self.send_header(
            headers = headers,
            version = version,
            code = code,
            code_s = code_s,
            stream = stream
        )

        # sends the part/payload information (data) to the client and optionally
        # flushes the current internal buffers to enforce sending of the value
        count += self.send_part(
            data,
            stream = stream,
            final = final,
            flush = flush,
            delay = delay,
            callback = callback
        )
        return count

    def send_header(
        self,
        headers = None,
        version = None,
        code = 200,
        code_s = None,
        stream = None,
        final = False,
        delay = True,
        callback = None
    ):
        # retrieves the various parts that define the response
        # and runs a series of normalization processes to retrieve
        # the relevant information of the data to be sent to client
        headers = headers or dict()
        version = version or "HTTP/1.1"
        code_s = code_s or netius.common.CODE_STRINGS.get(code, None)

        # creates the buffer list that is going to hold the complete set of
        # lines for the headers and then appends the complete set of headers
        # according to the previous construction
        buffer = []
        buffer.append("%s %d %s\r\n" % (version, code, code_s))
        for key, value in netius.legacy.iteritems(headers):
            key = netius.common.header_up(key)
            if not isinstance(value, list): value = (value,)
            for _value in value: buffer.append("%s: %s\r\n" % (key, _value))
        buffer.append("\r\n")
        buffer_data = "".join(buffer)

        # sends the buffer data to the connection peer so that it gets notified
        # about the headers for the current communication/message
        count = self.send_plain(
            buffer_data,
            stream = stream,
            delay = delay,
            callback = callback
        )

        # "notifies" the owner of the connection that the headers have been
        # sent all the http header information should be present
        self.owner.on_send_http(
            self.connection_ctx,
            self.parser_ctx,
            headers = headers,
            version = version,
            code = code,
            code_s = code_s
        )

        # returns the final number of bytes that have been sent during the current
        # operation of sending headers to the other peer
        return count

    def send_part(
        self,
        data,
        stream = None,
        final = True,
        flush = False,
        delay = True,
        callback = None
    ):
        if flush: count = self.send_base(data); self.flush(callback = callback)
        else: count = self.send_base(data, delay = delay, callback = callback)
        return count

    def parse(self, data):
        return self.parser.parse(data)

    def resolve_encoding(self, parser):
        # in case the "target" encoding is the plain one nothing
        # is required to be done as this is allowed by any kind
        # of http compliant connection (returns immediately)
        if self.encoding == PLAIN_ENCODING:
            self.current = PLAIN_ENCODING

        # if the target encoding is chunked must verify if the
        # type of http connection in question is 1.1 or above
        # otherwise must down-grade the encoding to plain
        elif self.encoding == CHUNKED_ENCODING:
            if parser.version < netius.common.HTTP_11:
                self.current = PLAIN_ENCODING
            else:
                self.current = CHUNKED_ENCODING

        # if the target encoding is the compressed gzip style the
        # current request must be verified have the version larger
        # than the 1.1 and the gzip encoding must be accepted
        elif self.encoding == GZIP_ENCODING:
            if parser.version < netius.common.HTTP_11:
                self.current = PLAIN_ENCODING
            elif not "gzip" in parser.get_encodings():
                self.current = CHUNKED_ENCODING
            else:
                self.current = GZIP_ENCODING

        # if the target encoding is the compressed deflate style the
        # current request must be verified have the version larger
        # than the 1.1 and the deflate encoding must be accepted
        elif self.encoding == DEFLATE_ENCODING:
            if parser.version < netius.common.HTTP_11:
                self.current = PLAIN_ENCODING
            elif not "deflate" in parser.get_encodings():
                self.current = CHUNKED_ENCODING
            else:
                self.current = DEFLATE_ENCODING

    def set_encoding(self, encoding):
        self.current = encoding

    def set_uncompressed(self):
        if self.current >= CHUNKED_ENCODING:
            self.current = CHUNKED_ENCODING
        else: self.current = PLAIN_ENCODING

    def set_plain(self):
        self.set_encoding(PLAIN_ENCODING)

    def set_chunked(self):
        self.set_encoding(CHUNKED_ENCODING)

    def set_gzip(self):
        self.set_encoding(GZIP_ENCODING)

    def set_deflate(self):
        self.set_encoding(DEFLATE_ENCODING)

    def is_plain(self):
        return self.current == PLAIN_ENCODING

    def is_chunked(self):
        return self.current > PLAIN_ENCODING

    def is_gzip(self):
        return self.current == GZIP_ENCODING

    def is_deflate(self):
        return self.current == DEFLATE_ENCODING

    def is_compressed(self):
        return self.current > CHUNKED_ENCODING

    def is_uncompressed(self):
        return not self.is_compressed()

    def is_flushed(self):
        return self.current > PLAIN_ENCODING

    def is_measurable(self, strict = True):
        if self.is_compressed(): return False
        if strict and self.is_chunked(): return False
        return True

    def on_data(self):
        self.owner.on_data_http(self.connection_ctx, self.parser_ctx)

    @contextlib.contextmanager
    def ctx_request(self, args = None, kwargs = None):
        yield

    @property
    def connection_ctx(self):
        return self

    @property
    def parser_ctx(self):
        return self.parser

    def _flush_plain(self, stream = None, callback = None):
        if not callback: return
        self.send_plain(b"", stream = stream, callback = callback)

    def _flush_chunked(self, stream = None, callback = None):
        self.send_plain(b"0\r\n\r\n", stream = stream, callback = callback)

    def _flush_gzip(self, stream = None, callback = None):
        # tries to retrieve a possible gzip object for the current
        # stream, note that no gzip object is going to be created
        # in case there's none defined for the current stream
        gzip = self._get_gzip(stream, ensure = False)

        # in case the gzip structure has not been initialized
        # (no data sent) no need to run the flushing of the
        # gzip data, so only the chunked part is flushed
        if not gzip:
            self._flush_chunked(stream = stream, callback = callback)
            return

        # flushes the internal zlib buffers to be able to retrieve
        # the data pending to be sent to the client and then sends
        # it using the chunked encoding strategy
        data_c = gzip.flush(zlib.Z_FINISH)
        self.send_chunked(data_c, stream = stream, final = False)

        # resets the gzip values to the original ones so that new
        # requests will starts the information from the beginning
        self._unset_gzip(stream)

        # runs the flush operation for the underlying chunked encoding
        # layer so that the client is correctly notified about the
        # end of the current request (normal operation)
        self._flush_chunked(stream = stream, callback = callback)

    def _get_gzip(self, stream, level = 6, ensure = True):
        # tries to retrieve the proper gzip object for the requested
        # stream and in case there's one or if the ensure flag is set
        # the retrieved value is returned to the caller method
        gzip = self.gzip_m.get(stream, None)
        if gzip or not ensure: return gzip

        # in case this is the first sending a new compress object
        # is created with the requested compress level, notice that
        # the special deflate case is handled differently
        is_deflate = self.is_deflate()
        wbits = -zlib.MAX_WBITS if is_deflate else zlib.MAX_WBITS | 16
        gzip = zlib.compressobj(level, zlib.DEFLATED, wbits)

        # updates the gzip objects map with the gzip object that has
        # just been created for the target stream
        self.gzip_m[stream] = gzip
        return gzip

    def _set_gzip(self, stream, gzip):
        self.gzip_m[stream] = gzip

    def _unset_gzip(self, stream):
        del self.gzip_m[stream]

    def _close_gzip(self, safe = True):
        # in case the gzip object is not defined returns the control
        # to the caller method immediately (nothing to be done)
        if not self.gzip_m: return

        # saves the current gzip object map locally (releasing reference)
        # and then recreates a new empty dictionary on the other object
        gzip_m = self.gzip_m
        self.gzip_m = dict()

        # iterates over the complete set of gzip object to run the flush
        # operation over each of them, as expected for proper and final
        # memory release, note that an exception may be raised in the case
        # of a duplicated flush operation
        for gzip in gzip_m.values():
            try:
                # runs the flush operation for the the final finish stage
                # (note that an exception may be raised)
                gzip.flush(zlib.Z_FINISH)
            except Exception:
                # in case the safe flag is not set re-raises the exception
                # to the caller stack (as expected by the callers)
                if not safe: raise

class HTTPServer(netius.StreamServer):
    """
    Base class for serving of the HTTP protocol, should contain
    the basic utilities for handling an HTTP request including
    headers and read of data.
    """

    BASE_HEADERS = {
        "Server" : netius.IDENTIFIER
    }
    """ The map containing the complete set of headers
    that are meant to be applied to all the responses """

    def __init__(self, encoding = "plain", common_log = None, *args, **kwargs):
        netius.StreamServer.__init__(self, *args, **kwargs)
        self.encoding_s = encoding
        self.common_log = common_log
        self.dynamic = False
        self.common_file = None

    @classmethod
    def build_data(
        cls,
        text,
        url = None,
        trace = False,
        style = True,
        style_urls = [],
        encode = True,
        encoding = "utf-8"
    ):
        if url: return cls.build_iframe(
            text,
            url,
            style = style,
            encode = encode,
            encoding = encoding
        )
        else: return cls.build_text(
            text,
            trace = trace,
            style = style,
            encode = encode,
            encoding = encoding
        )

    @classmethod
    def build_text(
        cls,
        text,
        trace = False,
        style = True,
        style_urls = [],
        encode = True,
        encoding = "utf-8"
    ):
        data = "".join(
            cls._gen_text(
                text,
                trace = trace,
                style = style,
                style_urls = style_urls
            )
        )
        if encode: data = netius.legacy.bytes(
            data,
            encoding = encoding,
            force = True
        )
        return data

    @classmethod
    def build_iframe(
        cls,
        text,
        url,
        style = True,
        style_urls = [],
        encode = True,
        encoding = "utf-8"
    ):
        data = "".join(
            cls._gen_iframe(
                text,
                url,
                style = style,
                style_urls = style_urls
            )
        )
        if encode: data = netius.legacy.bytes(
            data,
            encoding = encoding,
            force = True
        )
        return data

    @classmethod
    def _gen_text(cls, text, trace = False, style = True, style_urls = []):
        for value in cls._gen_header(text, style = style, style_urls = style_urls):
            yield value

        yield "<body>"
        yield "<h1>%s</h1>" % text
        if trace:
            lines = traceback.format_exc().splitlines()
            yield "<hr/>"
            yield "<div class=\"traceback\">"
            for line in lines: yield "<div>%s</div>" % line
            yield "</div>"
        yield "<hr/>"
        yield "<span>"
        yield netius.IDENTIFIER
        yield "</span>"
        yield "</body>"

        for value in cls._gen_footer(): yield value

    @classmethod
    def _gen_iframe(cls, text, url, style = True, style_urls = []):
        for value in cls._gen_header(text, style = style, style_urls = style_urls):
            yield value

        yield "<body>"
        yield "<iframe src=\"%s\"></iframe>" % url
        yield "</body>"

        for value in cls._gen_footer(): yield value

    @classmethod
    def _gen_header(cls, title, meta = True, style = True, style_urls = []):
        yield "<!DOCTYPE html>"
        yield "<html>"
        yield "<head>"
        if meta:
            yield "<meta charset=\"utf-8\" />"
            yield "<meta name=\"viewport\" content=\"width=device-width, user-scalable=no, initial-scale=1, minimum-scale=1, maximum-scale=1\" />"
        yield "<title>%s</title>" % title
        if style:
            for value in cls._gen_style(): yield value
            for style_url in style_urls:
                yield "<link rel=\"stylesheet\" type=\"text/css\" href=\"%s\" />" % style_url
        yield "</head>"

    @classmethod
    def _gen_footer(cls):
        yield "</html>"

    @classmethod
    def _gen_style(cls):
        yield netius.common.BASE_STYLE

    def cleanup(self):
        netius.StreamServer.cleanup(self)

        if self.common_file: self.common_file.close()

    def info_dict(self, full = False):
        info = netius.StreamServer.info_dict(self, full = full)
        info.update(encoding_s = self.encoding_s)
        return info

    def on_data(self, connection, data):
        netius.StreamServer.on_data(self, connection, data)
        connection.parse(data)

    def on_serve(self):
        netius.StreamServer.on_serve(self)
        if self.env: self.encoding_s = self.get_env("ENCODING", self.encoding_s)
        if self.env: self.common_log = self.get_env("COMMON_LOG", self.common_log)
        if self.common_log: self.common_file = open(self.common_log, "wb+")
        self.encoding = ENCODING_MAP.get(self.encoding_s, PLAIN_ENCODING)
        self.info("Starting HTTP server with '%s' encoding ..." % self.encoding_s)
        if self.common_log: self.info("Logging with Common Log Format to '%s' ..." % self.common_log)

    def build_connection(self, socket, address, ssl = False):
        return HTTPConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            encoding = self.encoding
        )

    def on_data_http(self, connection, parser):
        is_debug = self.is_debug()
        is_debug and self._log_request(connection, parser)
        connection.resolve_encoding(parser)

    def on_send_http(
        self,
        connection,
        parser,
        headers = None,
        version = None,
        code = 200,
        code_s = None
    ):
        self.common_file and self._log_request(
            connection,
            parser,
            headers = headers,
            version = version,
            code = code,
            code_s = code_s,
            output = self._write_common,
            mode = "common"
        )

    def on_flush_http(self, connection, parser, encoding = None):
        self.debug(
            "Connection '%s' from '%s' flushed" %\
            (connection.id, self.name)
        )

    def authorize(self, connection, parser, auth = None, **kwargs):
        # determines the proper authorization method to be used
        # taking into account either the provided method or the
        # default one in case none is provided
        auth = auth or netius.PasswdAuth
        if hasattr(auth, "auth"): auth_method = auth.auth
        else: auth_method = auth.auth_i
        if hasattr(auth, "is_simple"): is_simple = auth.is_simple()
        else: is_simple = auth.is_simple_i()

        # constructs a dictionary that contains extra information
        # about the current connection/request that may be used
        # to further determine if the request is authorized
        kwargs = dict(
            connection = connection,
            parser = parser,
            host = connection.address[0],
            port = connection.address[1],
            headers = parser.headers
        )

        # in case the current authentication method is considered
        # simples (no "classic" username and password) the named
        # arguments dictionary is provided as the only input
        if is_simple: return auth_method(**kwargs)

        # retrieves the authorization tuple (username and password)
        # using the current parser and verifies if at least one of
        # them is defined in case it's not returns an invalid result
        username, password = self._authorization(parser)
        if not username and not password: return False

        # uses the provided username and password to run the authentication
        # process using the method associated with the authorization structure
        return auth_method(username, password, **kwargs)

    def _apply_all(
        self,
        parser,
        connection,
        headers,
        upper = True,
        normalize = False,
        replace = False
    ):
        if upper: self._headers_upper(headers)
        if normalize: self._headers_normalize(headers)
        self._apply_base(headers, replace = replace)
        self._apply_parser(parser, headers, replace = replace)
        self._apply_connection(connection, headers)

    def _apply_base(self, headers, replace = False):
        cls = self.__class__
        for key, value in netius.legacy.iteritems(cls.BASE_HEADERS):
            if not replace and key in headers: continue
            headers[key] = value

    def _apply_parser(self, parser, headers, replace = False):
        if not replace and "Connection" in headers: return
        if parser.keep_alive: headers["Connection"] = "keep-alive"
        else: headers["Connection"] = "close"

    def _apply_connection(self, connection, headers, strict = True):
        is_chunked = connection.is_chunked()
        is_gzip = connection.is_gzip()
        is_deflate = connection.is_deflate()
        is_compressed = connection.is_compressed()
        is_measurable = connection.is_measurable(strict = strict)
        has_length = "Content-Length" in headers
        has_ranges = "Accept-Ranges" in headers

        if is_chunked: headers["Transfer-Encoding"] = "chunked"
        if is_gzip: headers["Content-Encoding"] = "gzip"
        if is_deflate: headers["Content-Encoding"] = "deflate"

        if not is_measurable and has_length: del headers["Content-Length"]
        if is_compressed and has_ranges: del headers["Accept-Ranges"]

    def _headers_upper(self, headers):
        for key, value in netius.legacy.items(headers):
            key_u = netius.common.header_up(key)
            del headers[key]
            headers[key_u] = value

    def _headers_normalize(self, headers):
        for key, value in netius.legacy.items(headers):
            if not type(value) in (list, tuple): continue
            headers[key] = ";".join(value)

    def _authorization(self, parser):
        # retrieves the headers from the parser structure and uses
        # them to retrieve the authorization header value returning
        # an invalid value in case no header is defined
        headers = parser.headers
        authorization = headers.get("authorization", None)
        if not authorization: return None, None

        # splits the authorization token between the realm and the
        # token value (decoding it afterwards) and then unpacks the
        # token into the username and password components (for validation)
        _realm, token = authorization.split(" ", 1)
        token = base64.b64decode(token)
        token = netius.legacy.str(token)
        username, password = token.split(":", 1)

        # retrieves the "final" packed result containing both the username
        # and the password associated with the authorization
        return username, password

    def _write_common(self, message, encoding = "utf-8"):
        message = netius.legacy.bytes(
            message,
            encoding = encoding,
            force = True
        )
        self.common_file.write(message + b"\n")
        self.common_file.flush()

    def _log_request(self, connection, parser, *args, **kwargs):
        mode = kwargs.pop("mode", "basic")
        method = getattr(self, "_log_request_" + mode)
        return method(connection, parser, *args, **kwargs)

    def _log_request_basic(self, connection, parser, output = None):
        # runs the defaulting operation on the logger output
        # method so that the default logger output is used instead
        output = output or self.debug

        # unpacks the various values that are going to be part of
        # the log message to be printed in the debug
        is_tuple = type(connection.address) in (list, tuple)
        ip_address = connection.address[0] if is_tuple else connection.address
        method = parser.method.upper()
        path = parser.get_path()
        version_s = parser.version_s

        # creates the message from the complete set of components
        # that are part of the current message and then prints a
        # debug message with the contents of it
        message = "%s %s %s @ %s" % (method, path, version_s, ip_address)
        output(message)

    def _log_request_common(
        self,
        connection,
        parser,
        headers = None,
        version = None,
        code = 200,
        code_s = None,
        size_s = None,
        username = "frank",
        output = None
    ):
        # runs the defaulting operation on the logger output
        # method so that the default logger output is used instead
        output = output or self.debug

        # unpacks the various values that are going to be part of
        # the log message to be printed in the debug, these values
        # should come from either the connection or the parser
        is_tuple = type(connection.address) in (list, tuple)
        ip_address = connection.address[0] if is_tuple else connection.address
        method = parser.method.upper()
        path = parser.get_path()
        version_s = parser.version_s

        # creates some composite values taking into account the various
        # parameters that have bean provided
        date_s = datetime.datetime.utcnow().strftime("%d/%b/%Y:%H:%M:%S +0000")
        size_s = size_s or headers.get("Content-Length", "-1")

        # creates the complete message in the "Common Log Format" and then
        # prints a debug message with that same contents
        message = "%s %s %s [%s] \"%s %s %s\" %d %s" % (
            ip_address, "-", username, date_s, method, path, version_s, code, size_s
        )
        output(message)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import base64

import netius.common

INTIAL_STATE = 1

HELO_STATE = 2

AUTH_STATE = 3

SESSION_STATE = 4

CHUNK_SIZE = 16384
""" The size of the chunk to be used for the partial sending
of the message to the client, this value will affect the
memory used by the server and its network performance """

CAPABILITIES = (
    "TOP",
    "USER",
    "STLS"
)
""" The capabilities that are going to be exposed to the client
as the ones handled by the server, should only expose the ones
that are properly handled by the server """

AUTH_METHODS = (
    "PLAIN",
)
""" Authentication methods that are available to be "used" by
the client, should be mapped into the proper auth handlers """

class POPConnection(netius.Connection):

    def __init__(self, host = "pop.localhost", *args, **kwargs):
        netius.Connection.__init__(self, *args, **kwargs)
        self.parser = None
        self.host = host
        self.username = None
        self.token_buf = []
        self.count = 0
        self.byte_c = 0
        self.size = 0
        self.file = None
        self.sizes = ()
        self.keys = ()
        self.state = INTIAL_STATE

    def open(self, *args, **kwargs):
        netius.Connection.open(self, *args, **kwargs)
        if not self.is_open(): return
        self.parser = netius.common.POPParser(self)
        self.parser.bind("on_line", self.on_line)

    def close(self, *args, **kwargs):
        netius.Connection.close(self, *args, **kwargs)
        if not self.is_closed(): return
        if self.file: self.file.close(); self.file = None
        if self.parser: self.parser.destroy()

    def parse(self, data):
        if self.state == AUTH_STATE: self.on_user(data)
        else: return self.parser.parse(data)

    def send_pop(self, message = "", lines = (), status = "OK", delay = True, callback = None):
        status_s = "+" + status if status == "OK" else "-" + status
        base = "%s %s" % (status_s, message)
        data = base + "\r\n"
        if lines: data += "\r\n".join(lines) + "\r\n.\r\n"
        count = self.send(data, delay = delay, callback = callback)
        self.owner.debug(base)
        return count

    def ready(self):
        self.assert_s(INTIAL_STATE)
        message = "POP3 server ready <%s@%s>" % (netius.NAME, self.host)
        self.send_pop(message)
        self.state = HELO_STATE

    def starttls(self):
        def callback(connection):
            connection.upgrade(server = True)
        message = "go ahead"
        self.send_pop(message, callback = callback)
        self.state = HELO_STATE

    def capa(self):
        self.assert_s(HELO_STATE)
        message = "list follows"
        self.send_pop(message, lines = CAPABILITIES)
        self.state = HELO_STATE

    def auth(self):
        self.assert_s(HELO_STATE)
        message = "list follows"
        self.send_pop(message, lines = AUTH_METHODS)
        self.state = HELO_STATE

    def accept(self):
        self.assert_s(HELO_STATE)
        self.ok()
        self.token_buf = []
        self.state = AUTH_STATE

    def stat(self):
        self.owner.on_stat_pop(self)
        message = "%d %d" % (self.count, self.byte_c)
        self.send_pop(message)

    def list(self):
        self.owner.on_list_pop(self)
        message = "%d messages (%d octets)" % (self.count, self.byte_c)
        lines = []
        for index in range(self.count):
            size = self.sizes[index]
            line = "%d %d" % (index, size)
            lines.append(line)
        self.send_pop(message, lines = lines)

    def uidl(self):
        self.owner.on_uidl_pop(self)
        message = "%d messages (%d octets)" % (self.count, self.byte_c)
        lines = []
        for index in range(self.count):
            key = self.keys[index]
            line = "%d %s" % (index, key)
            lines.append(line)
        self.send_pop(message, lines = lines)

    def retr(self, index):
        def callback(connection):
            if not connection.file: return
            file = connection.file
            contents = file.read(CHUNK_SIZE)
            if contents: self.send(contents, callback = callback)
            else: self.send("\r\n.\r\n"); file.close(); connection.file = None
        self.owner.on_retr_pop(self, index)
        message = "%d octets" % self.size
        self.send_pop(message, callback = callback)
        callback(self)

    def dele(self, index):
        self.owner.on_dele_pop(self, index)
        message = "removed"
        self.send_pop(message)

    def bye(self):
        message = "bye"
        self.send_pop(message)

    def ok(self):
        message = "accepted"
        self.send_pop(message)

    def not_implemented(self):
        message = "not implemented"
        self.send_pop(message, status = "ERR")

    def on_line(self, code, message):
        # "joins" the code and the message part of the message into the base
        # string and then uses this value to print some debug information
        base = "%s %s" % (code, message)
        self.owner.debug(base)

        # calls the proper top level owner based line information handler that
        # should ignore any usages as the connection will take care of the proper
        # handling for the current connection
        self.owner.on_line_pop(self, code, message)

        # converts the provided code into a lower case value and then uses it
        # to create the problem name for the handler method to be used
        code_l = code.lower()
        method_n = "on_" + code_l

        # verifies if the method for the current code exists in case it
        # does not raises an exception indicating the problem with the
        # code that has just been received (probably erroneous)
        extists = hasattr(self, method_n)
        if not extists: raise netius.ParserError("Invalid code '%s'" % code)

        # retrieves the reference to the method that is going to be called
        # for the handling of the current line from the current instance and
        # then calls it with the provided message
        method = getattr(self, method_n)
        method(message)

    def on_stls(self, message):
        self.starttls()

    def on_capa(self, message):
        self.capa()

    def on_auth(self, message):
        if message: self.accept()
        else: self.auth()

    def on_stat(self, message):
        self.stat()

    def on_list(self, message):
        self.list()

    def on_uidl(self, message):
        self.uidl()

    def on_retr(self, message):
        index = int(message)
        self.retr(index)

    def on_dele(self, message):
        index = int(message)
        self.dele(index)

    def on_quit(self, message):
        self.bye()
        self.close(flush = True)

    def on_user(self, token):
        # adds the partial token value to the token buffer and
        # then verifies if it contains the token if that's the
        # case continues the parsing otherwise returns immediately
        self.token_buf.append(token)
        index = token.find(b"\n")
        if index == -1: return

        # removes the extra characters from the token so that no
        # extra value is considered to be part of the token
        token = token.rstrip()

        # logs the received token as this handler is not a callback
        # from the typical on message method
        self.owner.debug(token)

        # joins the the token buffer recreating the token string
        # and then decodes into as a base 64 string and splits it
        # around its own components so that the proper auth callback
        # may be called to validate the authentication
        token = b"".join(self.token_buf)
        token_s = base64.b64decode(token)
        token_s = netius.legacy.str(token_s)
        _identifier, username, password = token_s.split("\0")

        # calls the callback to the authentication and in case everything
        # goes ok the ok value is called and the state of the current
        # connection is changed to session (authentication is valid)
        self.owner.on_auth_pop(self, username, password)
        self.ok()
        self.state = SESSION_STATE

    def assert_s(self, expected):
        if self.state == expected: return
        raise netius.ParserError("Invalid state")

class POPServer(netius.StreamServer):

    def __init__(self, adapter_s = "memory", auth_s = "dummy", *args, **kwargs):
        netius.StreamServer.__init__(self, *args, **kwargs)
        self.adapter_s = adapter_s
        self.auth_s = auth_s

    def serve(self, host = "pop.localhost", port = 110, *args, **kwargs):
        netius.StreamServer.serve(self, port = port, *args, **kwargs)
        self.host = host

    def on_connection_c(self, connection):
        netius.StreamServer.on_connection_c(self, connection)
        connection.ready()

    def on_connection_d(self, connection):
        netius.StreamServer.on_connection_d(self, connection)
        if connection.file: connection.file.close(); connection.file = None

    def on_data(self, connection, data):
        netius.StreamServer.on_data(self, connection, data)
        connection.parse(data)

    def on_serve(self):
        netius.StreamServer.on_serve(self)
        if self.env: self.host = self.get_env("POP_HOST", self.host)
        if self.env: self.adapter_s = self.get_env("POP_ADAPTER", self.adapter_s)
        if self.env: self.auth_s = self.get_env("POP_AUTH", self.auth_s)
        self.adapter = self.get_adapter(self.adapter_s)
        self.auth = self.get_auth(self.auth_s)
        self.info(
            "Starting POP server on '%s' using '%s' and '%s' ..." %
            (self.host, self.adapter_s, self.auth_s)
        )

    def build_connection(self, socket, address, ssl = False):
        return POPConnection(
            owner = self,
            socket = socket,
            address = address,
            ssl = ssl,
            host = self.host
        )

    def on_line_pop(self, connection, code, message):
        pass

    def on_auth_pop(self, connection, username, password):
        self.auth.auth_assert(username, password)
        connection.username = username

    def on_stat_pop(self, connection):
        username = connection.username
        count = self.adapter.count(owner = username)
        total = self.adapter.total(owner = username)
        connection.count = count
        connection.byte_c = total

    def on_list_pop(self, connection):
        username = connection.username
        sizes = self.adapter.sizes(owner = username)
        connection.sizes = sizes

    def on_uidl_pop(self, connection):
        username = connection.username
        connection.keys = self.adapter.list(owner = username)

    def on_retr_pop(self, connection, index):
        key = connection.keys[index]
        connection.size = self.adapter.size(key)
        connection.file = self.adapter.get_file(key)

    def on_dele_pop(self, connection, index):
        username = connection.username
        key = connection.keys[index]
        self.adapter.delete(key, owner = username)

if __name__ == "__main__":
    import logging
    server = POPServer(level = logging.DEBUG)
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import xml.dom.minidom

import netius.clients

def upnp_map(ext_port, int_port, host, protocol = "TCP", description = "netius"):
    """
    Defines a router port forwarding rule using an UPnP based
    request that tries to find the first available router.

    In case there's no available router with UPnP features the
    client may become idle, leaking memory.

    :see: http://www.upnp.org/specs/gw/UPnP-gw-WANIPConnection-v1-Service.pdf
    """

    message = """<?xml version="1.0"?>
        <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
            s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
            <s:Body>
                <u:AddPortMapping xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">
                    <NewRemoteHost></NewRemoteHost>
                    <NewExternalPort>%d</NewExternalPort>
                    <NewProtocol>%s</NewProtocol>
                    <NewInternalPort>%d</NewInternalPort>
                    <NewInternalClient>%s</NewInternalClient>
                    <NewEnabled>1</NewEnabled>
                    <NewPortMappingDescription>%s</NewPortMappingDescription>
                    <NewLeaseDuration>0</NewLeaseDuration>
                </u:AddPortMapping>
            </s:Body>
    </s:Envelope>""" % (ext_port, protocol, int_port, host, description)

    def on_location(connection, parser, request):
        data = request["data"]
        document = xml.dom.minidom.parseString(data)
        nodes = document.getElementsByTagName("controlURL")
        path = nodes[0].childNodes[0].data
        base_url = "http://%s:%d" % connection.address
        url = base_url + path
        netius.clients.HTTPClient.post_s(
            url,
            headers = dict(
                SOAPACTION = "\"urn:schemas-upnp-org:service:WANIPConnection:1#AddPortMapping\""
            ),
            data = message,
            asynchronous = False
        )
        client = connection.owner
        client.close()

    def on_headers(client, parser, headers):
        location = headers.get("Location", None)
        if not location: raise netius.DataError("No location found")
        http_client = netius.clients.HTTPClient()
        http_client.get(location, on_result = on_location)
        client.close()

    client = netius.clients.SSDPClient()
    client.bind("headers", on_headers)
    client.discover("urn:schemas-upnp-org:device:InternetGatewayDevice:1")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

import netius

class Stream(object):

    def open(self, mode = "r+b"):
        raise netius.NotImplemented("Missing implementation")

    def close(self):
        raise netius.NotImplemented("Missing implementation")

    def seek(self, offset):
        raise netius.NotImplemented("Missing implementation")

    def read(self, size):
        raise netius.NotImplemented("Missing implementation")

    def write(self, data):
        raise netius.NotImplemented("Missing implementation")

    def flish(self):
        raise netius.NotImplemented("Missing implementation")

class FileStream(Stream):

    def __init__(self, path, size):
        Stream.__init__(self)
        self.path = path
        self.size = size
        self.file = None

    def open(self, mode = "w+b", allocate = True):
        self.file = open(self.path, mode)
        if not allocate: return
        self.file.seek(self.size - 1)
        self.file.write(b"\0")
        self.file.flush()

    def close(self):
        if not self.file: return
        self.file.close()
        self.file = None

    def seek(self, offset):
        self.file.seek(offset)

    def read(self, size):
        return self.file.read(size)

    def write(self, data):
        self.file.write(data)

    def flush(self):
        self.file.flush()

class FilesStream(Stream):

    def __init__(self, dir_path, size, files_m):
        Stream.__init__(self)
        self.dir_path = dir_path
        self.size = size
        self.files_m = files_m
        self.files = []
        self._offset = 0

    def open(self, mode = "w+b", allocate = True):
        for file_m in self.files_m:
            file_path = file_m["path"]
            file_size = file_m["length"]
            file_path = os.path.join(self.dir_path, *file_path)
            file = open(file_path, mode)
            file_t = (file, file_m)
            self.files.append(file_t)
            if not allocate: continue
            file.seek(file_size - 1)
            file.write(b"\0")
            file.flush()

    def close(self):
        if not self.files: return
        for file_t in self.files:
            file, _file_m = file_t
            file.close()
        del self.files[:]

    def seek(self, offset):
        self._offset = offset

    def read(self, size):
        # starts a series of local variables that are going to
        # be used to control the offsets for multiple file stream
        # operations, these values have local meaning
        file_offset = 0
        offset = self._offset
        pending = size
        data_l = []

        # iterates over the complete set of file to read the
        # partial contents from each of the corresponding files
        # note that a data chunk may span multiple files
        for file_t in self.files:
            # unpacks the file tuple into the file stream and
            # the meta information map, and uses it to retrieve
            # the total size in bytes for the current file
            file, file_m = file_t
            file_size = file_m["length"]

            # calculates the possible start offset of the data
            # chunk and verifies that it's valid, less that the
            # size of the current file, otherwise skips the current
            # iteration, must go further
            start = offset - file_offset
            file_offset += file_size
            if start >= file_size: continue

            # calculates the end internal offset value as the
            # minimum value between the file size and the start
            # offset plus pending number of bytes, then uses
            # this end offset value to calculate the total number
            # of bytes to be read from the current file
            end = min(file_size, start + pending)
            count = end - start

            # seeks the current file to the internal start offset
            # value and reads the partial data from the stream,
            # adding then the chunk to the data (chunks) lists
            file.seek(start)
            chunk = file.read(count)
            data_l.append(chunk)

            # updates the pending (bytes) and offset values to the
            # new values so that the new iteration is coherent
            pending -= count
            offset += count

            # verifies if there's no more data pending and if
            # that's the case break the current loop as no more
            # files are going to be affected
            if pending == 0: break

        # updates the current offset of the (virtual) file stream
        # with length of the data that has been read, then avoids
        # overflows by truncating the value to the size
        self._offset += size
        self._offset = min(self.size, self._offset)

        # joins the complete set of data chunks in the data list
        # and returns the (final) buffer to the caller method
        return b"".join(data_l)

    def write(self, data):
        # starts a series of local variables that are going to
        # be used to control the offsets for multiple file stream
        # operations, these values have local meaning
        file_offset = 0
        offset = self._offset
        data_l = len(data)
        pending = data_l

        # iterates over the complete set of file to write the
        # partial contents to each of the corresponding files
        # note that a data chunk may span multiple files
        for file_t in self.files:
            # unpacks the file tuple into the file stream and
            # the meta information map, and uses it to retrieve
            # the total size in bytes for the current file
            file, file_m = file_t
            file_size = file_m["length"]

            # calculates the possible start offset of the data
            # chunk and verifies that it's valid, less that the
            # size of the current file, otherwise skips the current
            # iteration, must go further
            start = offset - file_offset
            file_offset += file_size
            if start >= file_size: continue

            # calculates the end internal offset value as the
            # minimum value between the file size and the start
            # offset plus pending number of bytes, then uses
            # this end offset value to calculate the total number
            # of bytes to be written to the current file
            end = min(file_size, start + pending)
            count = end - start

            # seeks the current file to the internal start offset
            # value and writes the partial data to the stream,
            # flushing then the file (avoiding corruption)
            file.seek(start)
            file.write(data[:count])
            file.flush()

            # updates the data chunk with the remaining data
            # taking into account the written amount of bytes
            # and updates the pending (bytes) and offset values
            data = data[count:]
            pending -= count
            offset += count

            # verifies if there's no more data pending and if
            # that's the case break the current loop as no more
            # files are going to be affected
            if pending == 0: break

        # updates the current offset of the (virtual) file stream
        # with length of the data that has just been written, then
        # avoids overflows by truncating the value to the size
        self._offset += data_l
        self._offset = min(self.size, self._offset)

    def flush(self):
        for file_t in self.files:
            file, _file_m = file_t
            file.flush()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import dhcp
from . import echo_ws
from . import echo
from . import ftp
from . import http
from . import http2
from . import mjpg
from . import pop
from . import proxy
from . import smtp
from . import socks
from . import tftp
from . import torrent
from . import ws
from . import wsgi

from .dhcp import DHCPRequest, DHCPServer
from .echo_ws import EchoWSServer
from .echo import EchoServer
from .ftp import FTPConnection, FTPServer
from .http import HTTPConnection, HTTPServer
from .http2 import HTTP2Server
from .mjpg import MJPGServer
from .pop import POPConnection, POPServer
from .proxy import ProxyConnection, ProxyServer
from .smtp import TERMINATION_SIZE, SMTPConnection, SMTPServer
from .socks import SOCKSConnection, SOCKSServer
from .tftp import TFTPRequest, TFTPServer
from .torrent import Pieces, TorrentTask, TorrentServer
from .ws import WSConnection, WSServer
from .wsgi import WSGIServer



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import copy

import netius.common
import netius.servers

class DHCPServerS(netius.servers.DHCPServer):

    def __init__(self, pool = None, options = {}, *args, **kwargs):
        netius.servers.DHCPServer.__init__(self, *args, **kwargs)

        self.pool = pool or netius.common.AddressPool("192.168.0.61", "192.168.0.69")
        self.options = {}
        self.lease = 3600

        self._build(options)

    def get_type(self, request):
        type = request.get_type()
        requested = request.get_requested()
        mac = request.get_mac()

        if type == 0x01: result = netius.common.OFFER_DHCP
        elif type == 0x03:
            current = self.pool.assigned(mac) or requested
            is_owner = self.pool.is_owner(mac, current)
            if is_owner: result = netius.common.ACK_DHCP
            else: result = netius.common.NAK_DHCP

        return result

    def get_options(self, request):
        options = copy.copy(self.options)
        return options

    def get_yiaddr(self, request):
        type = request.get_type()
        if type == 0x01: yiaddr = self._reserve(request)
        elif type == 0x03: yiaddr = self._confirm(request)
        return yiaddr

    def _build(self, options):
        lease = options.get("lease", {})
        self.lease = lease.get("time", 3600)

        for key, value in netius.legacy.iteritems(options):
            key_i = netius.common.OPTIONS_DHCP.get(key, None)
            if not key_i: continue
            self.options[key_i] = value

    def _reserve(self, request):
        mac = request.get_mac()
        return self.pool.reserve(
            owner = mac,
            lease = self.lease
        )

    def _confirm(self, request):
        requested = request.get_requested()
        mac = request.get_mac()
        current = self.pool.assigned(mac) or requested
        is_valid = self.pool.is_valid(current)
        if is_valid: self.pool.touch(current, self.lease)
        return current

if __name__ == "__main__":
    import logging
    host = netius.common.host()
    pool = netius.common.AddressPool("172.16.0.80", "172.16.0.89")
    options = dict(
        router = dict(routers = ["172.16.0.6"]),
        subnet = dict(subnet = "255.255.0.0"),
        dns = dict(
            servers = ["172.16.0.11", "172.16.0.12"]
        ),
        identifier = dict(identifier = host),
        broadcast = dict(broadcast = "172.16.255.255"),
        name = dict(name = "hive"),
        lease = dict(time = 3600),
        renewal = dict(time = 1800),
        rebind = dict(time = 2700),
        proxy = dict(url = "http://172.16.0.25:8080/proxy.pac")
    )
    server = DHCPServerS(
        pool = pool,
        options = options,
        level = logging.INFO
    )
    server.serve(env = True)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import select

POLL_TIMEOUT = 0.25
""" The timeout to be used under the all the poll methods
this should be considered the maximum amount of time a
thread waits for a poll request """

class Poll(object):
    """
    The top level abstract implementation of a poll object
    should be used for inheritance and reference on the
    various methods that are part of the api.
    """

    def __init__(self):
        self._open = False
        self.timeout = POLL_TIMEOUT
        self.read_o = {}
        self.write_o = {}
        self.error_o = {}

    @classmethod
    def name(cls):
        name = cls.__name__
        name = name[:-4]
        name = name.lower()
        return name

    @classmethod
    def test(cls):
        return True

    def open(self, timeout = POLL_TIMEOUT):
        if self._open: return
        self._open = True
        self.timeout = timeout

        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def close(self):
        if not self._open: return
        self._open = False

        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def poll(self):
        return []

    def poll_owner(self):
        reads, writes, errors = self.poll()

        result = dict()

        for read in reads:
            base = self.read_o[read]
            value = result.get(base, None)
            if not value:
                value = ([], [], [])
                result[base] = value
            value[0].append(read)

        for write in writes:
            base = self.write_o[write]
            value = result.get(base, None)
            if not value:
                value = ([], [], [])
                result[base] = value
            value[1].append(write)

        for error in errors:
            base = self.error_o[error]
            value = result.get(base, None)
            if not value:
                value = ([], [], [])
                result[base] = value
            value[2].append(error)

        return result

    def is_open(self):
        return self._open

    def is_edge(self):
        return False

    def is_empty(self):
        return not self.read_o and not self.write_o and not self.error_o

    def sub_all(self, socket, owner = None):
        self.sub_read(socket, owner = owner)
        self.sub_write(socket, owner = owner)
        self.sub_error(socket, owner = owner)

    def unsub_all(self, socket):
        self.unsub_error(socket)
        self.unsub_write(socket)
        self.unsub_read(socket)

    def is_sub_read(self, socket):
        return socket in self.read_o

    def is_sub_write(self, socket):
        return socket in self.write_o

    def is_sub_error(self, socket):
        return socket in self.error_o

    def sub_read(self, socket, owner = None):
        if socket in self.read_o: return
        self.read_o[socket] = owner

    def sub_write(self, socket, owner = None):
        if socket in self.write_o: return
        self.write_o[socket] = owner

    def sub_error(self, socket, owner = None):
        if socket in self.error_o: return
        self.error_o[socket] = owner

    def unsub_read(self, socket):
        if not socket in self.read_o: return
        del self.read_o[socket]

    def unsub_write(self, socket):
        if not socket in self.write_o: return
        del self.write_o[socket]

    def unsub_error(self, socket):
        if not socket in self.error_o: return
        del self.error_o[socket]

class EpollPoll(Poll):

    def __init__(self, *args, **kwargs):
        Poll.__init__(self, *args, **kwargs)
        self._open = False

    @classmethod
    def test(cls):
        return hasattr(select, "epoll")

    def open(self, timeout = POLL_TIMEOUT):
        if self._open: return
        self._open = True
        self.timeout = timeout

        self.epoll = select.epoll() #@UndefinedVariable

        self.fd_m = {}

        self.read_o = {}
        self.write_o = {}
        self.error_o = {}

    def close(self):
        if not self._open: return
        self._open = False

        for fd in self.fd_m: self.epoll.unregister(fd)
        self.epoll.close()
        self.epoll = None

        self.fd_m.clear()

        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def poll(self):
        result = ([], [], [])

        events = self.epoll.poll(self.timeout)
        for fd, event in events:
            if event & select.EPOLLIN: #@UndefinedVariable
                socket = self.fd_m.get(fd, None)
                socket and result[0].append(socket)
            if event & select.EPOLLOUT: #@UndefinedVariable
                socket = self.fd_m.get(fd, None)
                socket and result[1].append(socket)
            if event & select.EPOLLERR or event & select.EPOLLHUP:  #@UndefinedVariable
                socket = self.fd_m.get(fd, None)
                socket and result[2].append(socket)

        return result

    def is_edge(self):
        return True

    def sub_read(self, socket, owner = None):
        if socket in self.read_o: return
        socket_fd = socket.fileno()
        self.fd_m[socket_fd] = socket
        self.read_o[socket] = owner
        self.write_o[socket] = owner
        self.error_o[socket] = owner
        self.epoll.register( #@UndefinedVariable
            socket_fd,
            select.EPOLLIN | select.EPOLLOUT | select.EPOLLERR | select.EPOLLHUP | select.EPOLLET #@UndefinedVariable
        )

    def sub_write(self, socket, owner = None):
        pass

    def sub_error(self, socket, owner = None):
        pass

    def unsub_read(self, socket):
        if not socket in self.read_o: return
        socket_fd = socket.fileno()
        self.epoll.unregister( #@UndefinedVariable
            socket_fd
        )
        del self.fd_m[socket_fd]
        del self.read_o[socket]
        del self.write_o[socket]
        del self.error_o[socket]

    def unsub_write(self, socket):
        pass

    def unsub_error(self, socket):
        pass

class KqueuePoll(Poll):

    def __init__(self, *args, **kwargs):
        Poll.__init__(self, *args, **kwargs)
        self._open = False

    @classmethod
    def test(cls):
        return hasattr(select, "kqueue")

    def open(self, timeout = POLL_TIMEOUT):
        if self._open: return
        self._open = True
        self.timeout = timeout
        if self.timeout < 0: self.timeout = None

        self.kqueue = select.kqueue() #@UndefinedVariable

        self.fd_m = {}

        self.read_o = {}
        self.write_o = {}
        self.error_o = {}

    def close(self):
        if not self._open: return
        self._open = False

        self.kqueue.close()
        self.kqueue = None

        self.fd_m.clear()

        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def poll(self):
        result = ([], [], [])

        events = self.kqueue.control(None, 32, self.timeout)
        for event in events:
            if event.flags & select.KQ_EV_ERROR: #@UndefinedVariable
                socket = self.fd_m.get(event.udata, None)
                socket and result[2].append(socket)
            elif event.filter == select.KQ_FILTER_READ: #@UndefinedVariable
                socket = self.fd_m.get(event.udata, None)
                index = 2 if event.flags & select.KQ_EV_EOF else 0 #@UndefinedVariable
                socket and result[index].append(socket)
            elif event.filter == select.KQ_FILTER_WRITE: #@UndefinedVariable
                socket = self.fd_m.get(event.udata, None)
                index = 2 if event.flags & select.KQ_EV_EOF else 1 #@UndefinedVariable
                socket and result[index].append(socket)

        return result

    def is_edge(self):
        return True

    def sub_read(self, socket, owner = None):
        if socket in self.read_o: return
        socket_fd = socket.fileno()
        self.fd_m[socket_fd] = socket
        self.read_o[socket] = owner
        self.write_o[socket] = owner
        self.error_o[socket] = owner
        event = select.kevent( #@UndefinedVariable
            socket_fd,
            filter = select.KQ_FILTER_READ, #@UndefinedVariable
            flags = select.KQ_EV_ADD | select.KQ_EV_CLEAR, #@UndefinedVariable
            udata = socket_fd
        )
        self.kqueue.control([event], 0)
        event = select.kevent( #@UndefinedVariable
            socket_fd,
            filter = select.KQ_FILTER_WRITE, #@UndefinedVariable
            flags = select.KQ_EV_ADD | select.KQ_EV_CLEAR, #@UndefinedVariable
            udata = socket_fd
        )
        self.kqueue.control([event], 0)

    def sub_write(self, socket, owner = None):
        pass

    def sub_error(self, socket, owner = None):
        pass

    def unsub_read(self, socket):
        if not socket in self.read_o: return
        socket_fd = socket.fileno()
        event = select.kevent( #@UndefinedVariable
            socket_fd,
            filter = select.KQ_FILTER_READ, #@UndefinedVariable
            flags = select.KQ_EV_DELETE #@UndefinedVariable
        )
        self.kqueue.control([event], 0)
        event = select.kevent( #@UndefinedVariable
            socket_fd,
            filter = select.KQ_FILTER_WRITE, #@UndefinedVariable
            flags = select.KQ_EV_DELETE #@UndefinedVariable
        )
        self.kqueue.control([event], 0)
        del self.fd_m[socket_fd]
        del self.read_o[socket]
        del self.write_o[socket]
        del self.error_o[socket]

    def unsub_write(self, socket):
        pass

    def unsub_error(self, socket):
        pass

class PollPoll(Poll):

    def __init__(self, *args, **kwargs):
        Poll.__init__(self, *args, **kwargs)
        self._open = False

    @classmethod
    def test(cls):
        return hasattr(select, "poll")

    def open(self, timeout = POLL_TIMEOUT):
        if self._open: return
        self._open = True
        self.timeout = timeout

        self._poll = select.poll() #@UndefinedVariable

        self.read_fd = {}
        self.write_fd = {}

        self.read_o = {}
        self.write_o = {}
        self.error_o = {}

    def close(self):
        if not self._open: return
        self._open = False

        for fd in self.read_fd: self._poll.unregister(fd)
        self._poll = None

        self.read_fd.clear()
        self.write_fd.clear()

        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def poll(self):
        result = ([], [], [])

        events = self._poll.poll(self.timeout * 1000)
        for fd, event in events:
            if event & select.POLLIN: #@UndefinedVariable
                socket = self.read_fd.get(fd, None)
                socket and result[0].append(socket)
            if event & select.POLLOUT: #@UndefinedVariable
                socket = self.write_fd.get(fd, None)
                socket and result[1].append(socket)
            if event & select.POLLERR or event & select.POLLHUP: #@UndefinedVariable
                socket = self.read_fd.get(fd, None)
                socket and result[2].append(socket)

        return result

    def is_edge(self):
        return False

    def sub_read(self, socket, owner = None):
        if socket in self.read_o: return
        socket_fd = socket.fileno()
        self.read_fd[socket_fd] = socket
        self.read_o[socket] = owner
        self._poll.register( #@UndefinedVariable
            socket_fd,
            select.POLLIN #@UndefinedVariable
        )

    def sub_write(self, socket, owner = None):
        if socket in self.write_o: return
        socket_fd = socket.fileno()
        self.write_fd[socket_fd] = socket
        self.write_o[socket] = owner
        self._poll.modify( #@UndefinedVariable
            socket_fd,
            select.POLLIN | select.POLLOUT #@UndefinedVariable
        )

    def sub_error(self, socket, owner = None):
        if socket in self.error_o: return
        self.error_o[socket] = owner

    def unsub_read(self, socket):
        if not socket in self.read_o: return
        socket_fd = socket.fileno()
        self._poll.unregister( #@UndefinedVariable
            socket_fd
        )
        del self.read_fd[socket_fd]
        del self.read_o[socket]

    def unsub_write(self, socket):
        if not socket in self.write_o: return
        socket_fd = socket.fileno()
        self._poll.modify( #@UndefinedVariable
            socket_fd,
            select.POLLIN #@UndefinedVariable
        )
        del self.write_fd[socket_fd]
        del self.write_o[socket]

    def unsub_error(self, socket):
        if not socket in self.error_o: return
        del self.error_o[socket]

class SelectPoll(Poll):

    def __init__(self, *args, **kwargs):
        Poll.__init__(self, *args, **kwargs)
        self._open = False

    def open(self, timeout = POLL_TIMEOUT):
        if self._open: return
        self._open = True
        self.timeout = timeout
        if self.timeout < 0: self.timeout = None

        self.read_l = []
        self.write_l = []
        self.error_l = []

        self.read_o = {}
        self.write_o = {}
        self.error_o = {}

    def close(self):
        if not self._open: return
        self._open = False

        # removes the contents of all of the loop related structures
        # so that no extra selection operations are issued
        del self.read_l[:]
        del self.write_l[:]
        del self.error_l[:]

        # removes the complete set of elements from the map that associated
        # a socket with the proper owner
        self.read_o.clear()
        self.write_o.clear()
        self.error_o.clear()

    def poll(self):
        # "calculates" the amount of time the select method is going
        # to be sleeping for empty polls based on the fact that the
        # current timeout value may be unset
        sleep_timeout = self.timeout or POLL_TIMEOUT

        # verifies if the current selection list is empty
        # in case it's sleeps for a while and then continues
        # the loop (this avoids error in empty selection)
        is_empty = self.is_empty()
        if is_empty: time.sleep(sleep_timeout); return ([], [], [])

        # runs the proper select statement waiting for the desired
        # amount of time as timeout at the end a tuple with three
        # list for the different operations should be returned
        return select.select(
            self.read_l,
            self.write_l,
            self.error_l,
            self.timeout
        )

    def is_edge(self):
        return False

    def sub_read(self, socket, owner = None):
        if socket in self.read_o: return
        self.read_o[socket] = owner
        self.read_l.append(socket)

    def sub_write(self, socket, owner = None):
        if socket in self.write_o: return
        self.write_o[socket] = owner
        self.write_l.append(socket)

    def sub_error(self, socket, owner = None):
        if socket in self.error_o: return
        self.error_o[socket] = owner
        self.error_l.append(socket)

    def unsub_read(self, socket):
        if not socket in self.read_o: return
        self.read_l.remove(socket)
        del self.read_o[socket]

    def unsub_write(self, socket):
        if not socket in self.write_o: return
        self.write_l.remove(socket)
        del self.write_o[socket]

    def unsub_error(self, socket):
        if not socket in self.error_o: return
        self.error_l.remove(socket)
        del self.error_o[socket]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys
import time
import datetime
import threading

import netius

from .base import Middleware

class AnnoyerMiddleware(Middleware):
    """
    Simple middleware that prints an "annoying" status message
    to the standard output (stdout) from time to time providing
    a simple diagnostics strategy.
    """

    def __init__(self, owner, period = 10.0):
        Middleware.__init__(self, owner)
        self.period = period
        self._initial = None
        self._thread = None
        self._running = False

    def start(self):
        Middleware.start(self)
        self.period = netius.conf("ANNOYER_PERIOD", self.period, cast = float)
        self._thread = threading.Thread(target = self._run)
        self._thread.start()

    def stop(self):
        Middleware.stop(self)
        if self._thread:
            self._running = False
            self._thread.join()
            self._thread = None

    def _run(self):
        self._initial = datetime.datetime.utcnow()
        self._running = True
        while self._running:
            delta = datetime.datetime.utcnow() - self._initial
            delta_s = self.owner._format_delta(delta)
            message = "Uptime => %s | Connections => %d\n" %\
                (delta_s, len(self.owner.connections))
            sys.stdout.write(message)
            sys.stdout.flush()
            time.sleep(self.period)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from .conn import * #@UnusedWildImport
from .common import * #@UnusedWildImport

BUFFER_SIZE_S = None
""" The size of both the send and receive buffers for
the socket representing the server, this socket is
responsible for the handling of the new connections """

BUFFER_SIZE_C = None
""" The size of the buffers (send and receive) that
is going to be set on the on the sockets created by
the server (client sockets), this is critical for a
good performance of the server (large value) """

class Server(Base):

    def __init__(self, *args, **kwargs):
        Base.__init__(self, *args, **kwargs)
        self.receive_buffer_s = kwargs.get("receive_buffer_s", BUFFER_SIZE_S)
        self.send_buffer_s = kwargs.get("send_buffer_s", BUFFER_SIZE_S)
        self.receive_buffer_c = kwargs.get("receive_buffer_c", BUFFER_SIZE_C)
        self.send_buffer_c = kwargs.get("send_buffer_c", BUFFER_SIZE_C)
        self.socket = None
        self.host = None
        self.port = None
        self.type = None
        self.ssl = False
        self.key_file = None
        self.cer_file = None
        self.ca_file = None
        self.env = False
        self.allowed = []
        self._concrete = True

    def welcome(self):
        Base.welcome(self)

        self.info("Booting %s %s (%s) ..." % (NAME, VERSION, PLATFORM))

    def cleanup(self):
        Base.cleanup(self)

        # unsubscribes the current socket from all the positions in
        # the current polling mechanism, required for coherence
        self.unsub_all(self.socket)

        # tries to close the service socket, as this is the one that
        # has no connection associated and is independent
        try: self.socket and self.socket.close()
        except Exception: pass

        # unsets the socket attribute as the socket should now be closed
        # and not able to be used for any kind of communication
        self.socket = None

    def info_dict(self, full = False):
        info = Base.info_dict(self, full = full)
        info.update(
            host = self.host,
            port = self.port,
            type = self.type,
            ssl = self.ssl
        )
        return info

    def serve(
        self,
        host = None,
        port = 9090,
        type = TCP_TYPE,
        ipv6 = False,
        ssl = False,
        key_file = None,
        cer_file = None,
        ca_file = None,
        ca_root = True,
        ssl_verify = False,
        ssl_host = None,
        ssl_fingerprint = None,
        ssl_dump = False,
        setuid = None,
        backlog = socket.SOMAXCONN,
        load = True,
        start = True,
        env = False
    ):
        # processes the various default values taking into account if
        # the environment variables are meant to be processed for the
        # current context (default values are processed accordingly)
        host = self.get_env("HOST", host) if env else host
        port = self.get_env("PORT", port, cast = int) if env else port
        type = self.get_env("TYPE", type, cast = int) if env else type
        ipv6 = self.get_env("IPV6", ipv6, cast = bool) if env else ipv6
        ssl = self.get_env("SSL", ssl, cast = bool) if env else ssl
        port = self.get_env("UNIX_PATH", port) if env else port
        key_file = self.get_env("KEY_FILE", key_file) if env else key_file
        cer_file = self.get_env("CER_FILE", cer_file) if env else cer_file
        ca_file = self.get_env("CA_FILE", ca_file) if env else ca_file
        ca_root = self.get_env("CA_ROOT", ca_root, cast = bool) if env else ca_root
        ssl_verify = self.get_env("SSL_VERIFY", ssl_verify, cast = bool) if env else ssl_verify
        ssl_host = self.get_env("SSL_HOST", ssl_host) if env else ssl_host
        ssl_fingerprint = self.get_env("SSL_FINGERPRINT", ssl_fingerprint) if env else ssl_fingerprint
        ssl_dump = self.get_env("SSL_DUMP", ssl_dump) if env else ssl_dump
        key_file = self.get_env("KEY_DATA", key_file, expand = True) if env else key_file
        cer_file = self.get_env("CER_DATA", cer_file, expand = True) if env else cer_file
        ca_file = self.get_env("CA_DATA", ca_file, expand = True) if env else ca_file
        setuid = self.get_env("SETUID", setuid, cast = int) if env else setuid
        backlog = self.get_env("BACKLOG", backlog, cast = int) if env else backlog

        # runs the various extra variable initialization taking into
        # account if the environment variable is currently set or not
        # please note that some side effects may arise from this set
        if env: self.level = self.get_env("LEVEL", self.level)
        if env: self.diag = self.get_env("DIAG", self.diag, cast = bool)
        if env: self.middleware = self.get_env("MIDDLEWARE", self.middleware, cast = list)
        if env: self.children = self.get_env("CHILD", self.children, cast = int)
        if env: self.children = self.get_env("CHILDREN", self.children, cast = int)
        if env: self.logging = self.get_env("LOGGING", self.logging)
        if env: self.poll_name = self.get_env("POLL", self.poll_name)
        if env: self.poll_timeout = self.get_env(
            "POLL_TIMEOUT",
            self.poll_timeout,
            cast = float
        )
        if env: self.keepalive_timeout = self.get_env(
            "KEEPALIVE_TIMEOUT",
            self.keepalive_timeout,
            cast = int
        )
        if env: self.keepalive_interval = self.get_env(
            "KEEPALIVE_INTERVAL",
            self.keepalive_interval,
            cast = int
        )
        if env: self.keepalive_count = self.get_env(
            "KEEPALIVE_COUNT",
            self.keepalive_count,
            cast = int
        )
        if env: self.allowed = self.get_env("ALLOWED", self.allowed, cast = list)

        # updates the current service status to the configuration
        # stage as the next steps is to configure the service socket
        self.set_state(STATE_CONFIG)

        # starts the loading process of the base system so that the system should
        # be able to log some information that is going to be output
        if load: self.load()

        # ensures the proper default address value, taking into account
        # the type of connection that is currently being used, this avoids
        # problems with multiple stack based servers (ipv4 and ipv6)
        if host == None: host = "::1" if ipv6 else "127.0.0.1"

        # defaults the provided SSL key and certificate paths to the
        # ones statically defined (dummy certificates), please beware
        # that using these certificates may create validation problems
        key_file = key_file or SSL_KEY_PATH
        cer_file = cer_file or SSL_CER_PATH
        ca_file = ca_file or SSL_CA_PATH

        # populates the basic information on the currently running
        # server like the host the port and the (is) SSL flag to be
        # used latter for reference operations
        self.host = host
        self.port = port
        self.type = type
        self.ssl = ssl
        self.ssl_host = ssl_host
        self.ssl_fingerprint = ssl_fingerprint
        self.ssl_dump = ssl_dump
        self.env = env

        # populates the key, certificate and certificate authority file
        # information with the values that have just been resolved, these
        # values are going to be used for runtime certificate loading
        self.key_file = key_file
        self.cer_file = cer_file
        self.ca_file = ca_file

        # determines if the client side certificate should be verified
        # according to the loaded certificate authority values or if
        # on the contrary no (client) validation should be performed
        ssl_verify = ssl_verify or False

        # verifies if the type of server that is going to be created is
        # unix or internet based, this allows the current infra-structure
        # to work under the much more latency free unix sockets
        is_unix = host == "unix"

        # checks the type of service that is meant to be created and
        # creates a service socket according to the defined service
        family = socket.AF_INET6 if ipv6 else socket.AF_INET
        family = socket.AF_UNIX if is_unix else family
        if type == TCP_TYPE: self.socket = self.socket_tcp(
            ssl,
            key_file = key_file,
            cer_file = cer_file,
            ca_file = ca_file,
            ca_root = ca_root,
            ssl_verify = ssl_verify,
            family = family
        )
        elif type == UDP_TYPE: self.socket = self.socket_udp()
        else: raise errors.NetiusError("Invalid server type provided '%d'" % type)

        # "calculates" the address "bind target", taking into account that this
        # server may be running under a unix based socket infra-structure and
        # if that's the case the target (file path) is also removed, avoiding
        # a duplicated usage of the socket (required for address re-usage)
        address = port if is_unix else (host, port)
        if is_unix and os.path.exists(address): os.remove(address)

        # binds the socket to the provided address value (per spec) and then
        # starts the listening in the socket with the provided backlog value
        # defaulting to the typical maximum backlog as possible if not provided
        self.socket.bind(address)
        if type == TCP_TYPE: self.socket.listen(backlog)

        # in case the set user id value the user of the current process should
        # be changed so that it represents the new (possibly unprivileged user)
        if setuid: os.setuid(setuid)

        # in case the selected port is zero based, meaning that a randomly selected
        # port has been assigned by the bind operation the new port must be retrieved
        # and set for the current server instance as the new port (for future reference)
        if self.port == 0: self.port = self.socket.getsockname()[1]

        # creates the string that identifies it the current service connection
        # is using a secure channel (SSL) and then prints an info message about
        # the service that is going to be started
        ipv6_s = " on IPv6" if ipv6 else ""
        ssl_s = " using SSL" if ssl else ""
        self.info("Serving '%s' service on %s:%s%s%s ..." % (self.name, host, port, ipv6_s, ssl_s))

        # runs the fork operation responsible for the forking of the
        # current process into the various child processes for multiple
        # process based parallelism, note that this must be done after
        # the master socket has been created (to be shared), note that
        # in case the result is not valid an immediate return is performed
        # as this represents a master based process (not meant to serve)
        result = self.fork()
        if not result: return

        # ensures that the current polling mechanism is correctly open as the
        # service socket is going to be added to it next, this overrides the
        # default behavior of the common infra-structure (on start)
        self.poll = self.build_poll()
        self.poll.open(timeout = self.poll_timeout)

        # adds the socket to all of the pool lists so that it's ready to read
        # write and handle error, this is the expected behavior of a service
        # socket so that it can handle all of the expected operations
        self.sub_all(self.socket)

        # calls the on serve callback handler so that underlying services may be
        # able to respond to the fact that the service is starting and some of
        # them may print some specific debugging information
        self.on_serve()

        # starts the base system so that the event loop gets started and the
        # the servers get ready to accept new connections (starts service)
        if start: self.start()

    def socket_tcp(
        self,
        ssl = False,
        key_file = None,
        cer_file = None,
        ca_file = None,
        ca_root = True,
        ssl_verify = False,
        family = socket.AF_INET,
        type = socket.SOCK_STREAM
    ):
        # verifies if the provided family is of type internet and if that's
        # the case the associated flag is set to valid for usage
        is_inet = family in (socket.AF_INET, socket.AF_INET6)

        # retrieves the proper string based type for the current server socket
        # and the prints a series of log message about the socket to be created
        type_s = " SSL" if ssl else ""
        self.debug("Creating server's TCP%s socket ..." % type_s)
        if ssl: self.debug("Loading '%s' as key file" % key_file)
        if ssl: self.debug("Loading '%s' as certificate file" % cer_file)
        if ssl and ca_file: self.debug("Loading '%s' as certificate authority file" % ca_file)
        if ssl and ssl_verify: self.debug("Loading with client SSL verification")

        # creates the socket that it's going to be used for the listening
        # of new connections (server socket) and sets it as non blocking
        _socket = socket.socket(family, type)
        _socket.setblocking(0)

        # in case the server is meant to be used as SSL wraps the socket
        # in suck fashion so that it becomes "secured"
        if ssl: _socket = self._ssl_wrap(
            _socket,
            key_file = key_file,
            cer_file = cer_file,
            ca_file = ca_file,
            ca_root = ca_root,
            server = True,
            ssl_verify = ssl_verify
        )

        # sets the various options in the service socket so that it becomes
        # ready for the operation with the highest possible performance, these
        # options include the reuse address to be able to re-bind to the port
        # and address and the keep alive that drops connections after some time
        # avoiding the leak of connections (operative system managed)
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if is_inet: _socket.setsockopt(
            socket.IPPROTO_TCP,
            socket.TCP_NODELAY,
            1
        )
        if self.receive_buffer_s: _socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_RCVBUF,
            self.receive_buffer_s
        )
        if self.send_buffer_s: _socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_SNDBUF,
            self.send_buffer_s
        )
        self._socket_keepalive(_socket)

        # returns the created tcp socket to the calling method so that it
        # may be used from this point on
        return _socket

    def socket_udp(self, family = socket.AF_INET, type = socket.SOCK_DGRAM):
        # prints a small debug message about the udp socket that is going
        # to be created for the server's connection
        self.debug("Creating server's udp socket ...")

        # creates the socket that it's going to be used for the listening
        # of new connections (server socket) and sets it as non blocking
        _socket = socket.socket(family, type)
        _socket.setblocking(0)

        # sets the various options in the service socket so that it becomes
        # ready for the operation with the highest possible performance
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        # returns the created udp socket to the calling method so that it
        # may be used from this point on
        return _socket

    def on_serve(self):
        pass

class DatagramServer(Server):

    def __init__(self, *args, **kwargs):
        Server.__init__(self, *args, **kwargs)
        self.renable = True
        self.wready = True
        self.pending_s = 0
        self.pending = collections.deque()
        self.pending_lock = threading.RLock()

    def reads(self, reads, state = True):
        Server.reads(self, reads, state = state)
        for read in reads:
            self.on_read(read)

    def writes(self, writes, state = True):
        Server.writes(self, writes, state = state)
        for write in writes:
            self.on_write(write)

    def errors(self, errors, state = True):
        Server.errors(self, errors, state = state)
        for error in errors:
            self.on_error(error)

    def serve(self, type = UDP_TYPE, *args, **kwargs):
        Server.serve(self, type = type, *args, **kwargs)

    def on_read(self, _socket):
        # tries to retrieve a proper callback for the socket
        # that received the read operations and calls the
        # proper callback as expected
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("read", _socket)

        # in case the read enabled flag is not currently set
        # must return immediately because the read operation
        # is not currently being allowed
        if not self.renable == True: return

        # verifies if the provided socket for reading is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

        try:
            # iterates continuously trying to read as much data as possible
            # when there's a failure to read more data it should raise an
            # exception that should be handled properly, note that if the
            # read enabled flag changed in the middle of the read handler
            # the loop is stop as no more read operations are allowed
            while True:
                data, address = _socket.recvfrom(CHUNK_SIZE)
                self.on_data(address, data)
                if not self.renable == True: break
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception)

    def on_write(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("write", _socket)

        # verifies if the provided socket for writing is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

        try:
            self._send(_socket)
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception)

    def on_error(self, _socket):
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("error", _socket)

        # verifies if the provided socket for error is the same
        # as the one registered in the client if that's not the case
        # return immediately to avoid unwanted operations
        if not _socket == self.socket: return

    def on_exception(self, exception):
        self.warning(exception)
        self.log_stack()

    def on_expected(self, exception):
        self.debug(exception)

    def on_data(self, address, data):
        pass

    def ensure_write(self):
        # retrieves the identifier of the current thread and
        # checks if it's the same as the one defined in the
        # owner in case it's not then the operation is not
        # considered to be safe and must be delayed
        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.tid

        # in case the thread where this code is being executed
        # is not the same the operation is considered to be not
        # safe and so it must be delayed to be executed in the
        # next loop of the thread cycle, must return immediately
        # to avoid extra subscription operations
        if not is_safe: return self.delay(self.ensure_write, safe = True)

        # adds the current socket to the list of write operations
        # so that it's going to be available for writing as soon
        # as possible from the poll mechanism
        self.sub_write(self.socket)

    def remove_write(self):
        self.unsub_write(self.socket)

    def enable_read(self):
        if not self.renable == False: return
        self.renable = True
        self.sub_read(self.socket)

    def disable_read(self):
        if not self.renable == True: return
        self.renable = False
        self.unsub_read(self.socket)

    def send(self, data, address, delay = True, callback = None):
        data = legacy.bytes(data)
        data_l = len(data)

        if callback: data = (data, callback)
        data = (data, address)

        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.tid

        self.pending_lock.acquire()
        try: self.pending.appendleft(data)
        finally: self.pending_lock.release()

        self.pending_s += data_l

        if self.wready:
            if is_safe and not delay: self._flush_write()
            else: self.delay(
                self._flush_write,
                immediately = True,
                verify = True,
                safe = True
            )
        else:
            self.ensure_write()

    def _send(self, _socket):
        self.wready = True
        self.pending_lock.acquire()
        try:
            while True:
                # in case there's no pending data to be sent to the
                # client side breaks the current loop (queue empty)
                if not self.pending: break

                # retrieves the current data from the pending list
                # of data to be sent and then saves the original data
                # object (for latter usage), sets the callback as not
                # defined and then unpacks the data into data and address
                data = self.pending.pop()
                data_o = data
                callback = None
                data, address = data

                # verifies if the data type of the data is a tuple and
                # if that's the case unpacks it as data and callback
                is_tuple = type(data) == tuple
                if is_tuple: data, callback = data

                # retrieves the length (in bytes) of the data that is
                # going to be sent to the client
                data_l = len(data)

                try:
                    # tries to send the data through the socket and
                    # retrieves the number of bytes that were correctly
                    # sent through the socket, this number may not be
                    # the same as the size of the data in case only
                    # part of the data has been sent
                    if data: count = _socket.sendto(data, address)
                    else: count = 0

                    # verifies if the current situation is that of a non
                    # closed socket and valid data, and if that's the case
                    # and no data has been sent the socket is considered to
                    # be in a would block situation and and such an error
                    # is raised indicating the issue (is going to be caught
                    # as a normal would block exception)
                    if data and count == 0: raise socket.error(errno.EWOULDBLOCK)
                except:
                    # sets the current connection write ready flag to false
                    # so that a new level notification must be received
                    self.wready = False

                    # ensures that the write event is going to be triggered
                    # this is required so that the remaining pending data is
                    # going to be correctly written on a new write event,
                    # triggered when the connection is ready for more writing
                    self.ensure_write()

                    # in case there's an exception must add the data
                    # object to the list of pending data because the data
                    # has not been correctly sent
                    self.pending.append(data_o)
                    raise
                else:
                    # decrements the size of the pending buffer by the number
                    # of bytes that were correctly send through the buffer
                    self.pending_s -= count

                    # verifies if the data has been correctly sent through
                    # the socket and for suck situations calls the callback
                    # object, otherwise creates a new data object with only
                    # the remaining (partial data) and the callback to be
                    # sent latter (only then the callback is called)
                    is_valid = count == data_l
                    if is_valid:
                        callback and callback(self)
                    else:
                        data_o = ((data[count:], callback), address)
                        self.pending.append(data_o)
        finally:
            self.pending_lock.release()

        self.remove_write()

    def _flush_write(self):
        """
        Flush operations to be called by the delaying controller
        (in ticks) that will trigger all the write operations
        pending for the current connection's socket.
        """

        self.writes((self.socket,), state = False)

class StreamServer(Server):

    def reads(self, reads, state = True):
        Server.reads(self, reads, state = state)
        for read in reads:
            if read == self.socket: self.on_read_s(read)
            else: self.on_read(read)

    def writes(self, writes, state = True):
        Server.writes(self, writes, state = state)
        for write in writes:
            if write == self.socket: self.on_write_s(write)
            else: self.on_write(write)

    def errors(self, errors, state = True):
        Server.errors(self, errors, state = state)
        for error in errors:
            if error == self.socket: self.on_error_s(error)
            else: self.on_error(error)

    def serve(self, type = TCP_TYPE, *args, **kwargs):
        Server.serve(self, type = type, *args, **kwargs)

    def on_read_s(self, _socket):
        try:
            while True:
                socket_c, address = _socket.accept()
                try: self.on_socket_c(socket_c, address)
                except Exception: socket_c.close(); raise
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected_s(error)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception_s(error)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected_s(error)
            elif not error_v in VALID_ERRORS:
                self.on_exception_s(error)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception_s(exception)

    def on_write_s(self, _socket):
        pass

    def on_error_s(self, _socket):
        pass

    def on_read(self, _socket):
        # tries to retrieve the connection from the provided socket
        # object (using the associative map) this connection is going
        # to be used over the method for multiple operations
        connection = self.connections_m.get(_socket, None)

        # in case the connection does not inherit from the server
        # connection then the handling should be done by the upper
        # layer, meaning that the handling is done upstream
        if hasattr(connection, "_base"):
            return Base.on_read(self, _socket)

        # tries to retrieve a possible callback registered for the socket
        # and if there's one calls it to be able to "append" extra operations
        # to the execution of the read operation in the socket
        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("read", _socket)

        # run a series of validations, in case there no connection
        # or the connection is not ready for return the control flow is
        # returned to the caller method (nothing to be done)
        if not connection: return
        if not connection.status == OPEN: return
        if not connection.renable == True: return

        try:
            # verifies if there's any pending operations in the
            # connection (eg: SSL handshaking) and performs it trying
            # to finish them, if they are still pending at the current
            # state returns immediately (waits for next loop)
            if self._pending(connection): return

            # iterates continuously trying to read as much data as possible
            # when there's a failure to read more data it should raise an
            # exception that should be handled properly
            while True:
                data = connection.recv(CHUNK_SIZE)
                if data: self.on_data(connection, data)
                else: connection.close(); break
                if not connection.status == OPEN: break
                if not connection.renable == True: break
                if not connection.socket == _socket: break
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error, connection)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error, connection)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception, connection)

    def on_write(self, _socket):
        connection = self.connections_m.get(_socket, None)

        # in case the connection does not inherit from the server
        # connection then the handling should be done by the upper
        # layer, meaning that the handling is done upstream
        if hasattr(connection, "_base"):
            return Base.on_write(self, _socket)

        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("write", _socket)

        if not connection: return
        if not connection.status == OPEN: return

        try:
            connection._send()
        except ssl.SSLError as error:
            error_v = error.args[0] if error.args else None
            error_m = error.reason if hasattr(error, "reason") else None
            if error_v in SSL_SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in SSL_VALID_ERRORS and\
                not error_m in SSL_VALID_REASONS:
                self.on_exception(error, connection)
        except socket.error as error:
            error_v = error.args[0] if error.args else None
            if error_v in SILENT_ERRORS:
                self.on_expected(error, connection)
            elif not error_v in VALID_ERRORS:
                self.on_exception(error, connection)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException as exception:
            self.on_exception(exception, connection)

    def on_error(self, _socket):
        connection = self.connections_m.get(_socket, None)

        # in case the connection does not inherit from the server
        # connection then the handling should be done by the upper
        # layer, meaning that the handling is done upstream
        if hasattr(connection, "_base"):
            return Base.on_error(self, _socket)

        callbacks = self.callbacks_m.get(_socket, None)
        if callbacks:
            for callback in callbacks: callback("error", _socket)

        if not connection: return
        if not connection.status == OPEN: return

        connection.close()

    def on_exception(self, exception, connection):
        self.warning(exception)
        self.log_stack()
        connection.close()

    def on_exception_s(self, exception):
        self.warning(exception)
        self.log_stack()

    def on_expected(self, exception, connection):
        self.debug(exception)
        connection.close()

    def on_expected_s(self, exception):
        self.debug(exception)

    def on_upgrade(self, connection):
        connection.set_upgraded()

    def on_ssl(self, connection):
        # in case an SSL host verification value is defined for the server
        # the client connection is going to be verified against such host
        # to make sure the client represents the expected entity, note that
        # as a fallback the SSL verification process is performed with no
        # value defined, meaning that a possible (SSL) host value set in the
        # connection is going to be used instead for the verification
        if self.ssl_host: connection.ssl_verify_host(self.ssl_host)
        else: connection.ssl_verify_host()

        # in case the SSL fingerprint verification process is enabled for the
        # current server the client certificates are going to be verified for
        # their integrity using this technique, otherwise the default verification
        # process is going to be run instead
        if self.ssl_fingerprint: connection.ssl_verify_fingerprint(self.ssl_fingerprint)
        else: connection.ssl_verify_fingerprint()

        # in case the SSL dump flag is set the dump operation is performed according
        # to that flag, otherwise the default operation is performed, that in most
        # of the cases should prevent the dump of the information
        if self.ssl_dump: connection.ssl_dump_certificate(self.ssl_dump)
        else: connection.ssl_dump_certificate()

        # in case the current connection is under the upgrade
        # status calls the proper event handler so that the
        # connection workflow may proceed accordingly
        if connection.upgrading: self.on_upgrade(connection)

    def on_data(self, connection, data):
        connection.set_data(data)

    def on_socket_c(self, socket_c, address):
        # verifies if the current address (host value) is present in
        # the currently defined allowed list and in case that's not
        # the case raises an exception indicating the issue
        host = address[0] if address else ""
        result = netius.common.assert_ip4(host, self.allowed)
        if not result: raise errors.NetiusError(
            "Address '%s' not present in allowed list" % host
        )

        # verifies a series of pre-conditions on the socket so
        # that it's ensured to be in a valid state before it's
        # set as a new connection for the server (validation)
        if self.ssl and not socket_c._sslobj: socket_c.close(); return

        # in case the SSL mode is enabled, "patches" the socket
        # object with an extra pending reference, that is going
        # to be to store pending callable operations in it
        if self.ssl: socket_c.pending = None

        # verifies if the socket is of type internet (either ipv4
        # of ipv6), this is going to be used for conditional setting
        # of some of the socket options
        is_inet = socket_c.family in (socket.AF_INET, socket.AF_INET6)

        # sets the socket as non blocking and then updated a series
        # of options in it, some of them taking into account if the
        # socket if of type internet (timeout values)
        socket_c.setblocking(0)
        socket_c.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if is_inet: socket_c.setsockopt(
            socket.IPPROTO_TCP,
            socket.TCP_NODELAY,
            1
        )
        if self.receive_buffer_c: socket_c.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_RCVBUF,
            self.receive_buffer_c
        )
        if self.send_buffer_c: socket_c.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_SNDBUF,
            self.send_buffer_c
        )

        # the process creation is considered completed and a new
        # connection is created for it and opened, from this time
        # on a new connection is considered accepted/created for server
        connection = self.build_connection(socket_c, address, ssl = self.ssl)
        connection.open()

        # registers the SSL handshake method as a starter method
        # for the connection, so that the handshake is properly
        # performed on the initial stage of the connection (as expected)
        if self.ssl: connection.add_starter(self._ssl_handshake)

        # runs the initial try for the handshaking process, note that
        # this is an async process and further tries to the handshake
        # may come after this one (async operation) in case an exception
        # is raises the connection is closed (avoids possible errors)
        try: connection.run_starter()
        except Exception: connection.close(); raise

        # in case there's extraneous data pending to be read from the
        # current connection's internal receive buffer it must be properly
        # handled on the risk of blocking the newly created connection
        if connection.is_pending_data(): self.on_read(connection.socket)

    def on_socket_d(self, socket_c):
        connection = self.connections_m.get(socket_c, None)
        if not connection: return

    def _ssl_handshake(self, connection):
        Server._ssl_handshake(self, connection)

        # verifies if the socket still has finished the SSL handshaking
        # process (by verifying the appropriate flag) and then if that's
        # not the case returns immediately (nothing done)
        if not connection.ssl_handshake: return

        # prints a debug information notifying the developer about
        # the finishing of the handshaking process for the connection
        self.debug("SSL Handshaking completed for connection")

        # calls the proper callback on the connection meaning
        # that SSL is now enabled for that socket/connection and so
        # the communication between peers is now secured
        self.on_ssl(connection)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Netius System
# Copyright (c) 2008-2019 Hive Solutions Lda.
#
# This file is part of Hive Netius System.
#
# Hive Netius System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Netius System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Netius System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2019 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import ssl
import time
import uuid
import errno
import socket
import datetime
import threading
import collections

from . import tls
from . import config
from . import legacy
from . import observer

OPEN = 1
""" The open status value, meant to be used in situations
where the status of the entity is open (opposite of d) """

CLOSED = 2
""" Closed status value to be used in entities which have
no pending structured opened and operations are limited """

PENDING = 3
""" The pending status used for transient states (eg created)
connections under this state must be used carefully """

CHUNK_SIZE = 16384
""" The size of the chunk to be used while received
data from the service socket """

class BaseConnection(observer.Observable):
    """
    Abstract connection object that should encapsulate
    a socket object enabling it to be accessed in much
    more "protected" way avoiding possible sync problems.

    It should also abstract the developer from all the
    select associated complexities adding and removing the
    underlying socket from the selecting mechanism for the
    appropriate operations.
    """

    def __init__(
        self,
        owner = None,
        socket = None,
        address = None,
        datagram = False,
        ssl = False,
        max_pending = -1,
        min_pending = -1
    ):
        observer.Observable.__init__(self)
        self.status = PENDING
        self.id = str(uuid.uuid4())
        self.connecting = False
        self.upgrading = False
        self.owner = owner
        self.socket = socket
        self.address = address
        self.datagram = datagram
        self.ssl = ssl
        self.ssl_host = None
        self.ssl_fingerprint = None
        self.ssl_handshake = False
        self.ssl_connecting = False
        self.ssl_dump = False
        self.max_pending = max_pending
        self.min_pending = min_pending
        self.renable = True
        self.wready = False
        self.pending_s = 0
        self.restored_s = 0
        self.starters = collections.deque()
        self.pending = collections.deque()
        self.restored = collections.deque()
        self.pending_lock = threading.RLock()
        self.restored_lock = threading.RLock()
        self._starter = None

    def destroy(self):
        observer.Observable.destroy(self)
        self.pending_s = 0
        self.restored_s = 0
        self.pending.clear()
        self.restored.clear()

    def open(self, connect = False):
        # in case the current status of the connection is already open
        # it does not make sense to proceed with the opening of the
        # connection as the connection is already open
        if self.status == OPEN: return

        # retrieves the reference to the owner object from the
        # current instance to be used to add the socket to the
        # proper pooling mechanisms (at least for reading)
        owner = self.owner

        # registers the socket for the proper reading mechanisms
        # in the polling infra-structure of the owner
        owner.sub_read(self.socket)
        owner.sub_error(self.socket)

        # adds the current connection object to the list of
        # connections in the owner and the registers it in
        # the map that associates the socket with the connection
        owner.connections.append(self)
        owner.connections_m[self.socket] = self

        # sets the status of the current connection as open
        # as all the internal structures have been correctly
        # updated and not it's safe to perform operations
        self.status = OPEN

        # in case the connect flag is set must set the current
        # connection as connecting indicating that some extra
        # steps are still required to complete the connection
        if connect: self.set_connecting()

        # calls the top level on connection creation handler so that the owner
        # object gets notified about the creation of the connection (open)
        owner.new_connection(self)

        # triggers the open event in the current connection so that any listening
        # object is notified about the opening of this connection, as requested by
        # the current netius specification and strategy
        self.trigger("open", self)

    def close(self, flush = False, destroy = True):
        # in case the current status of the connection is closes it does
        # nor make sense to proceed with the closing as the connection
        # is already in the closed state (nothing to be done)
        if self.status == CLOSED: return

        # in case the flush flag is set a different approach is taken
        # where all the pending data is flushed (as possible) before
        # the connection is effectively closed, this is only valid in
        # case the current connection status is open and not connecting
        if flush and self.status == OPEN and not self.connecting:
            return self.close_flush()

        # immediately sets the status of the connection as closed
        # so that no one else changed the current connection status
        # this is relevant to avoid any erroneous situation
        self.status = CLOSED

        # unsets the connecting flag this is for connections that
        # are under the connecting state and that must be reverted
        # to the original not connecting state
        self.connecting = False

        # unsets the upgrading flag as the connection could not be
        # under the upgrading state anymore as the closing process
        # for the connection has been started (not ready for upgrade)
        self.upgrading = False

        # unsets the write ready flag so that no more write operations
        # are performed as requested by specification (mandatory) this
        # should avoid extra erroneous write operations
        self.wready = False

        # resets the size of the data pending to be send and the clears
        # the list of pending information (invalidation the previous one)
        self.pending_s = 0
        self.pending.clear()

        # resets the complete set of restored (to receive) data so that
        # no more data is set as pending to read (invalidation)
        self.restored_s = 0
        self.restored.clear()

        # retrieves the reference to the owner object from the
        # current instance to be used to removed the socket from the
        # proper pooling mechanisms (at least for reading)
        owner = self.owner

        # removes the socket from all the polling mechanisms so that
        # interaction with it is no longer part of the selecting mechanism
        self.owner.unsub_all(self.socket)

        # removes the current connection from the list of connection in the
        # owner and also from the map that associates the socket with the
        # proper connection (also in the owner)
        if self in owner.connections: owner.connections.remove(self)
        if self.socket in owner.connections_m: del owner.connections_m[self.socket]

        # closes the socket, using the proper gracefully way so that
        # operations are no longer allowed in the socket, in case there's
        # an error in the operation fails silently (on purpose)
        try: self.socket.close()
        except Exception: pass

        # calls the top level on connection delete handler so that the owner
        # object gets notified about the deletion of the connection (closed)
        owner.del_connection(self)

        # triggers the close event in the current connection so that any listening
        # object is notified about the closing of this connection, as requested by
        # the current netius specification and strategy
        self.trigger("close", self)

        # in case the destroy flag is set the current object is destroy meaning that
        # for instance the current event registered handlers will no longer be available
        # this is important to avoid any memory leak from circular references, from
        # this moment on the connection is considered disabled (not ready for usage)
        if destroy: self.destroy()

    def close_flush(self):
        self.send(None, callback = self._close_callback)

    def upgrade(self, key_file = None, cer_file = None, ca_file = None, server = True):
        # in case the current connection is already an SSL oriented one there's
        # nothing to be done here and the method returns immediately to caller
        if self.ssl: return

        # prints a debug message about the upgrading of the connection that is
        # going to be performed for the current connection
        self.owner.debug("Upgrading '%s' into an SSL connection ..." % self.id)

        # sets the SSL flag of the current connection as the connection is now
        # going to be considered as SSL oriented and then sets the upgrading flag
        # for the same connection so that the main event loop "knows" how to handle
        # new event on this connection properly
        self.ssl = True
        self.upgrading = True

        # determines if the arguments based certificate and key values should be used
        # of if instead the owner values should be used as a fallback process, these
        # values are going to be used as part of the SSL upgrade process
        if hasattr(self.owner, "key_file"): key_file = key_file or self.owner.key_file
        if hasattr(self.owner, "cer_file"): cer_file = cer_file or self.owner.cer_file
        if hasattr(self.owner, "ca_file"): ca_file = ca_file or self.owner.ca_file

        # prints some debug information about the files that are going to be used for
        # the SSL based connection upgrade, this is mainly for debugging purposes
        if key_file: self.owner.debug("Using '%s' as key file" % key_file)
        if cer_file: self.owner.debug("Using '%s' as certificate file" % cer_file)
        if ca_file: self.owner.debug("Using '%s' as certificate authority file" % ca_file)

        # removes the "old" association socket association for the connection and
        # unsubscribes the "old" socket from the complete set of events, this should
        # be enough to avoid any interaction with the "old" socket
        del self.owner.connections_m[self.socket]
        self.owner.unsub_all(self.socket)

        # upgrades the current socket into an SSL oriented socket, note that the server
        # flag controls if the socket to be created is a client or a server one this
        # is relevant for the SSL handshaking part of the connection, the resulting
        # encapsulated SSL socket is then set as the current connection's socket
        self.socket = self.owner._ssl_upgrade(
            self.socket,
            key_file = key_file,
            cer_file = cer_file,
            ca_file = ca_file,
            server = server
        )

        # updates the current socket in connection resolution map with the new SSL one
        # and then subscribes the same socket for the read and error events
        self.owner.connections_m[self.socket] = self
        self.owner.sub_read(self.socket)
        self.owner.sub_error(self.socket)

        # adds the SSL handshake method as a starter for the current connection (to be
        # called before read) and then runs the kickoff starter operation to start
        # the connection "upgrading" process (as expected)
        self.add_starter(self.owner._ssl_handshake)
        self.run_starter()

    def set_connecting(self):
        self.connecting = True
        self.ensure_write()

    def set_connected(self):
        self.remove_write()
        self.connecting = False
        self.trigger("connect", self)

    def set_upgraded(self):
        self.upgrading = False
        self.trigger("upgrade", self)

    def set_data(self, data, address = None):
        if address: self.trigger("data", self, data, address)
        else: self.trigger("data", self, data)

    def ensure_write(self, flush = True):
        # retrieves the identifier of the current thread and
        # checks if it's the same as the one defined in the
        # owner in case it's not then the operation is not
        # considered to be safe and must be delayed
        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.owner.tid

        # in case the thread where this code is being executed
        # is not the same the operation is considered to be not
        # safe and so it must be delayed to be executed in the
        # next loop of the thread cycle, must return immediately
        # to avoid extra subscription operations
        if not is_safe: return self.owner.delay(self.ensure_write, safe = True)

        # verifies if the status of the connection is open and
        # in case it's not returns immediately as there's no reason
        # to so it for writing
        if not self.status == OPEN: return

        # in case the write ready flag is enabled (writes allowed)
        # and the flush parameter is set the ensure operation performs
        # the flush of the write operations (instead of subscription)
        # this may be done because it's safe to flush the write operations
        # when the write ready flag for the connection is set
        if self.wready and flush: return self._flush_write()

        # verifies if the owner object is already subscribed for the
        # write operation in case it is returns immediately in order
        # avoid any extra subscription operation
        if self.owner.is_sub_write(self.socket): return

        # adds the current socket to the list of write operations
        # so that it's going to be available for writing as soon
        # as possible from the poll mechanism
        self.owner.sub_write(self.socket)

    def remove_write(self):
        if not self.status == OPEN: return
        self.owner.unsub_write(self.socket)

    def enable_read(self):
        """
        Enables read operations for the current connection
        this will set the read enable flag and then subscribe
        to the read operations in case the underlying poll
        method is not of type edge (level based).

        This is a dangerous operation as it may cause the system
        to stall if misused.
        """

        if not self.status == OPEN: return
        if not self.renable == False: return

        self.renable = True
        self.owner.sub_read(self.socket)

    def disable_read(self):
        """
        Disables any read operation on the current socket, does that
        by disabling the current read enable and then unsubscribing
        the current connection from the read operation.

        This is an extremely dangerous operation and the correct knowledge
        of the event poll is required to avoid stalling.
        """

        if not self.status == OPEN: return
        if not self.renable == True: return

        self.renable = False
        self.owner.unsub_read(self.socket)

    def send(self, data, address = None, delay = True, force = False, callback = None):
        """
        The main send call to be used by a proxy connection and
        from different threads.

        This method is the equivalent on a socket basis to both the
        send and the send to method, meaning that datagram send operations
        are also allowed by using the address parameter.

        In case the sending should be forced as delayed (next tick)
        the delay flag should be set and the sending will be delayed.
        This is especially useful to avoid a stack overflow situation
        because of extended callback calling, for example while sending
        very large chunks of information (eg: multi megabyte files).

        An optional callback attribute may be sent, so that when the
        send is complete it's called with a reference to the data object.

        Calling this method should be done with care as this can
        create dead lock or socket corruption situations, extreme
        knowledge of the internals of the system is required.

        :type data: String
        :param data: The buffer containing the data to be sent
        through this connection to the other endpoint.
        :type address: Tuple
        :param address: The target address for the send operation,
        this is relevant only for datagram based connections.
        :type delay: bool
        :param delay: If the send operation should be delayed until
        the next tick operation or if it should be performed as
        soon as possible (as defined in specification).
        :type force: bool
        :param force: If the sending of the data should be "forced",
        meaning that even if the connection is not open the data
        is added to the current pending queue. Useful for client
        connections wanting to write ahead.
        :type callback: Function
        :param callback: Function to be called when the data set
        to be send is completely sent to the socket.
        """

        # ensures that the data type of the current data string
        # is the required one for the output operations (binary)
        # in case it's not the required transformation operations
        # should be performed so that the data format is compatible
        data = legacy.bytes(data) if data else data

        # calculates the size in bytes of the provided data so
        # that it may be used latter for the incrementing of
        # of the total size of pending bytes
        data_l = len(data) if data else 0

        # verifies that the connection is currently in the open
        # state and then verifies if that's not the case returns
        # immediately, not possible to send data
        if not self.status == OPEN and not force: return 0

        # creates the tuple that is going to represent the data
        # to be sent, this tuple should contain the data itself
        # the target address (for datagram) and the callback
        data = (data, address, callback)

        # retrieves the identifier of the current thread and then
        # verifies if it's the same as thread where the event loop
        # is being executed (safe execution) for options to be taken
        cthread = threading.current_thread()
        tid = cthread.ident or 0
        is_safe = tid == self.owner.tid

        # runs the pend operation that adds the current data to the
        # structures that control the data pending in output to be
        # sent in the proper (flush) write operation
        self.pend(data)

        # verifies if the write ready flag is set, for that case the
        # write flushing operation must be performed, so that all pending
        # bytes to be sent in the connection may be flushed as fast as
        # possible (avoiding extra loops in polling)
        if self.wready:
            # checks if the safe flag is set and if it is runs
            # the send operation right way otherwise "waits" until
            # the next tick operation (delayed execution), note that
            # running the flush operation immediately may lead to
            # typical stack overflow errors (due to recursion limit)
            if is_safe and not delay: self._flush_write()
            else: self.owner.delay(
                self._flush_write,
                immediately = True,
                verify = True,
                safe = True
            )

        # otherwise the write stream is not ready and so the
        # connection must be ensured to be write ready, should
        # subscribe to the write events as soon as possible
        else: self.ensure_write()

        # returns the final number of bytes (length of data)
        # that has been submitted to be sent (as soon as possible)
        return data_l

    def recv(self, size = CHUNK_SIZE, force = False):
        if not self.status == OPEN and not force: return b""
        return self._recv(size = size)

    def pend(self, data, back = True):
        # verifies if the provided data is a tuple and if that's
        # the case unpacks the callback value from it, required
        is_tuple = type(data) == tuple
        if is_tuple: data_b, _address, _callback = data
        else: data_b = data

        # calculates the size in bytes of the provided data so
        # that it may be used latter for the incrementing of
        # of the total size of pending bytes
        data_l = len(data_b) if data_b else 0

        # acquires the pending lock and then inserts the data into
        # the list of pending information to sent to the client end
        # point, notice that it's inserted at the beginning of the list
        # as the pop operation is performed at the end of list, meaning
        # that the fifo strategy is maintained
        self.pending_lock.acquire()
        try:
            if back: self.pending.appendleft(data)
            else: self.pending.append(data)
        finally:
            self.pending_lock.release()

        # increments the size of the pending data to be sent by
        # the size of the inner data buffer to be added (as requested)
        self.pending_s += data_l

        # sends the pend event indicating that a new set of data has been
        # set as pending in the internal buffers of the connection and
        # some of the flow controlling operation may have to be performed
        self.trigger("pend", self)

    def restore(self, data, back = True):
        """
        Restore data to the pending (to receive) so that they are
        going to be "received" in the next receive operation.

        :type data: String
        :param data: The buffer of data that is going to be restored
        to the internal receive buffers.
        :type back: bool
        :param back: If the data should be restore to the "back" of
        the internal buffers, or if instead it should be added to the
        front (next to be received) part of the buffer.
        """

        # calculates the size in bytes of the provided data so
        # that it may be used latter for the incrementing of
        # of the total size of restored bytes
        data_l = len(data) if data else 0

        # acquires the restored lock (thread safety) and then adds
        # the sequence of data to the restored buffer so that it's
        # going to be used in the next receive operation
        self.restored_lock.acquire()
        try:
            if back: self.restored.appendleft(data)
            else: self.restored.append(data)
        finally:
            self.restored_lock.release()

        # increments the size of the restored count by the size of
        # the data that has just been restored
        self.restored_s += data_l

    def run_starter(self):
        # iterates continuously around the set of registered
        # starters for the current connection to try to run as
        # much as possible all of their logic, some of them may
        # fail as the complete set of resources are not available
        # and should resume on next loop tick
        while True:
            # tries to retrieve the proper starter method to be
            # used in the current iteration, either from the currently
            # pending operation, next in line or as fallback an invalid
            # one, that is going to invalidate the iteration
            if self._starter: starter = self._starter
            elif self.starters: starter = self.starters.pop()
            else: starter = None

            # in case there's no starter pending and no other is set
            # in the queue for the connection (to be executed next)
            # breaks the current loop (nothing left to be done)
            if not starter: break

            # sets the current starter as the starter currently selected
            # by the loop set of operations
            self._starter = starter
            starter(self)

            # tries to determine if the start has finished and if
            # that's note the case breaks the loop, as stater should
            # finish on the next loop tick
            finished = self._starter == None
            if not finished: return True

        # returns the default invalid value, meaning that no more starter
        # operations are pending for the the current connection
        return False

    def end_starter(self):
        self._starter = None

    def add_starter(self, starter, back = True):
        if back: self.starters.appendleft(starter)
        else: self.starters.append(starter)

    def remove_starter(self, starter):
        self.starters.remove(starter)

    def info_dict(self, full = False):
        info = dict(
            status = self.status,
            id = self.id,
            connecting = self.connecting,
            upgrading = self.upgrading,
            address = self.address,
            ssl = self.ssl,
            renable = self.renable,
            wready = self.wready,
            pending_s = self.pending_s,
            restored_s = self.restored_s
        )
        return info

    def ssl_certificate(self, binary = False):
        if not self.ssl: return None
        return self.socket.getpeercert(binary_form = binary)

    def ssl_verify_host(self, host = None):
        host = host or self.ssl_host
        if not host: return
        certificate = self.ssl_certificate()
        tls.match_hostname(certificate, host)

    def ssl_verify_fingerprint(self, fingerprint = None):
        fingerprint = fingerprint or self.ssl_fingerprint
        if not fingerprint: return
        certificate = self.ssl_certificate(binary = True)
        tls.match_fingerprint(certificate, fingerprint)

    def ssl_dump_certificate(self, dump = False):
        dump = dump or self.ssl_dump
        if not dump: return
        certificate = self.ssl_certificate()
        certificate_binary = self.ssl_certificate(binary = True)
        tls.dump_certificate(certificate, certificate_binary)

    def ssl_protocol(self):
        return self.ssl_alpn_protocol() or self.ssl_npn_protocol()

    def ssl_alpn_protocol(self):
        if not self.socket: return None
        if not hasattr(self.socket, "selected_alpn_protocol"): return None
        return self.socket.selected_alpn_protocol()

    def ssl_npn_protocol(self):
        if not self.socket: return None
        if not hasattr(self.socket, "selected_npn_protocol"): return None
        return self.socket.selected_npn_protocol()

    def is_open(self):
        return self.status == OPEN

    def is_closed(self):
        return self.status == CLOSED

    def is_pending(self):
        return self.status == PENDING

    def is_connected(self):
        return not self.connecting

    def is_connecting(self):
        return self.connecting

    def is_upgrading(self):
        return self.upgrading

    def is_throttleable(self):
        return True

    def is_exhausted(self):
        return self.pending_s > self.max_pending

    def is_restored(self):
        return self.pending_s <= self.min_pending

    def is_pending_data(self):
        return self.restored_s > 0

    def _send(self):
        # sets the write ready flag so that any further request to
        # write operation will be immediately performed
        self.wready = True

        # in case the current connection is under the connecting stage
        # it's not possible to perform a sending operation and the
        # send operation is ignored, note that the write ready flag
        # is still set as it may be used later for flushing operations
        if self.connecting: return

        # acquires the pending lock so that no other access to the
        # the pending structure is made from a different thread
        self.pending_lock.acquire()

        try:
            # iterates continuously so that all the pending data to be
            # sent is correctly sent to the other peer if that's possible
            while True:
                # verifies if there's data pending to be sent in case
                # there's not returns immediately, because there's
                # nothing pending to be done for such case
                if not self.pending: break

                # retrieves the current data chunk to be send from the
                # list of pending things and then saves the data chunk
                # in an "original" object an tries to unpack it in case
                # the type of it is a tuple
                data = self.pending.pop()
                data_o = data
                data, address, callback = data
                is_close = data == None
                data_l = 0 if is_close else len(data)

                try:
                    # tries to send the data through the socket and
                    # retrieves the number of bytes that were correctly
                    # sent through the socket, this number may not be
                    # the same as the size of the data in case only
                    # part of the data has been sent, note that if no
                    # data is provided the shutdown operation is performed
                    # instead to close the stream between both sockets
                    if is_close: self._shutdown(); count = 0
                    elif address: count = self.socket.sendto(data, address)
                    elif data: count = self.socket.send(data)
                    else: count = 0

                    # verifies if the current situation is that of a non
                    # closed socket and valid data, and if that's the case
                    # and no data has been sent the socket is considered to
                    # be in a would block situation and and such an error
                    # is raised indicating the issue (is going to be caught
                    # as a normal would block exception)
                    if not is_close and data and count == 0:
                        raise socket.error(errno.EWOULDBLOCK)
                except:
                    # sets the current connection write ready flag to false
                    # so that a new level notification must be received
                    self.wready = False

                    # ensures that the write event is going to be triggered
                    # this is required so that the remaining pending data is
                    # going to be correctly written on a new write event,
                    # triggered when the connection is ready for more writing
                    self.ensure_write()

                    # in case there's an exception must re-add the data
                    # object to the list of pending data because the data
                    # has not been correctly sent
                    self.pending.append(data_o)

                    # re-raises the current to the upper layers so that they
                    # can properly handle what happened here, sometime this
                    # is going to imply a graceful handling other times the
                    # exception is critical and a more severe handling applies
                    raise
                else:
                    # decrements the size of the pending buffer by the number
                    # of bytes that were correctly send through the buffer
                    self.pending_s -= count

                    # verifies if the data has been correctly (and completely)
                    # sent through the socket and if that's not the case re-pushes
                    # the pending data back to the pending stack
                    is_valid = count == data_l
                    if not is_valid:
                        data_o = (data[count:], address, callback)
                        self.pending.append(data_o)

                    # triggers the unpend event as some of the data has been
                    # removed from the pending buffer and so any listener must
                    # be notified so that flow operations may be performed
                    self.trigger("unpend", self)

                    # in case the is valid flag is set (all of the data for
                    # the current write operation has been sent) calls the
                    # the associated callback (in case it exists)
                    if is_valid and callback: callback(self)
        finally:
            # releases the pending access lock so that no leaks
            # exists and no access to the pending is prevented
            self.pending_lock.release()

        # removes the current connection from the set of connections
        # that are monitored for any write event (no longer required)
        self.remove_write()

    def _recv(self, size):
        data = self._recv_restored(size)
        if data: return data
        if self.datagram: return self.socket.recvfrom(size)
        else: return self.socket.recv(size)

    def _recv_ssl(self, size):
        data = self._recv_restored(size)
        if data: return data
        has_socket = hasattr(self.socket, "_sock")
        if has_socket: return self.socket._sock.recv(size)
        else: return socket.socket.recv(self.socket, size)

    def _recv_restored(self, size):
        if not self.restored_s: return b""
        data = self.restored.pop()
        data = data[:size]
        remaining = data[size:]
        self.restored_s -= len(data)
        if remaining: self.restore(remaining, back = False)
        return data

    def _shutdown(self, close = False, force = False, ignore = True):
        # in case the status of the current connection is
        # already closed returns immediately as it's not
        # possible to shutdown a closed connection
        if self.status == CLOSED: return

        try:
            # verifies the type of connection and taking that
            # into account runs the proper shutdown operation
            # either the SSL based shutdown that unwraps the
            # current secure connection and send a graceful
            # shutdown notification to the other peer, or the
            # normal shutdown operation for the socket
            if self.ssl and hasattr(self.socket._sslobj, "shutdown"):
                self.socket._sslobj.shutdown()
            if force: self.socket.shutdown(socket.SHUT_RDWR)
        except (IOError, socket.error, ssl.SSLError):
            # ignores the IO/SSL error that has just been raise, this
            # assumes that the problem that has just occurred is not
            # relevant as the socket is shutting down and if a problem
            # occurs that must be related with the socket being closed
            # on the other side of the connection
            if not ignore: raise

        # in case the close (connection) flag is active the
        # current connection should be closed immediately
        # following the successful shutdown of the socket
        if close: self.close()

    def _close_callback(self, connection):
        """
        The callback to the delayed (flush based) close operation
        of the connection. This callback should be able to destroy
        and close all the resources associated with the connection.

        :type connection: Connection
        :param connection: The connection associated with the callback
        that is being called, this connection will be closed.
        """

        connection.close()

    def _flush_write(self):
        """
        Flush operations to be called by the delaying controller
        (in ticks) that will trigger all the write operations
        pending for the current connection's socket.
        """

        self.owner.writes((self.socket,), state = False)

class DiagConnection(BaseConnection):

    def __init__(self, *args, **kwargs):
        BaseConnection.__init__(self, *args, **kwargs)
        self.creation = time.time()
        self.recvs = 0
        self.sends = 0
        self.in_bytes = 0
        self.out_bytes = 0

    def recv(self, *args, **kwargs):
        result = BaseConnection.recv(self, *args, **kwargs)
        self.in_bytes += len(result) if result else 0
        self.recvs += 1
        return result

    def send(self, data, *args, **kwargs):
        result = BaseConnection.send(self, data, *args, **kwargs)
        self.out_bytes += result
        self.sends += 1
        return result

    def info_dict(self, full = False):
        info = BaseConnection.info_dict(self, full = full)
        info.update(
            uptime = self._uptime(),
            recvs = self.recvs,
            sends = self.sends,
            in_bytes = self.in_bytes,
            out_bytes = self.out_bytes
        )
        geo = self._resolve(self.address)
        if geo: info["geo"] = geo
        return info

    def _uptime(self):
        creation_d = datetime.datetime.utcfromtimestamp(self.creation)
        delta = datetime.datetime.utcnow() - creation_d
        delta_s = self.owner._format_delta(delta)
        return delta_s

    def _resolve(self, address):
        import netius.common
        ip, _port = address
        return netius.common.GeoResolver.resolve(ip)

is_diag = config.conf("DIAG", False, cast = bool)
if is_diag: Connection = DiagConnection
else: Connection = BaseConnection
